from infrahub.core.diff.repository.deserializer import EnrichedDiffDeserializer
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffDeserializerDependency(DependencyBuilder[EnrichedDiffDeserializer]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> EnrichedDiffDeserializer:
        return EnrichedDiffDeserializer()



from pydantic import Field

from infrahub.message_bus.messages.proposed_change.base_with_diff import BaseProposedChangeWithDiffMessage


class RequestProposedChangeDataIntegrity(BaseProposedChangeWithDiffMessage):
    """Sent trigger data integrity checks for a proposed change"""


class RequestProposedChangeRunGenerators(BaseProposedChangeWithDiffMessage):
    """Sent trigger the generators that are impacted by the proposed change to run."""

    refresh_artifacts: bool = Field(..., description="Whether to regenerate artifacts after the generators are run")
    do_repository_checks: bool = Field(
        ..., description="Whether to run repository and user checks after the generators are run"
    )


class RequestProposedChangeRepositoryChecks(BaseProposedChangeWithDiffMessage):
    """Sent when a proposed change is created to trigger additional checks"""


class RequestProposedChangeSchemaIntegrity(BaseProposedChangeWithDiffMessage):
    """Sent trigger schema integrity checks for a proposed change"""


class RequestProposedChangeUserTests(BaseProposedChangeWithDiffMessage):
    """Sent trigger to run tests (smoke, units, integrations) for a proposed change."""



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, Literal, overload

from .exceptions import NodeNotFoundError

if TYPE_CHECKING:
    from .client import SchemaType
    from .node import InfrahubNode, InfrahubNodeSync


def get_schema_name(schema: str | type[SchemaType] | None = None) -> str | None:
    if isinstance(schema, str):
        return schema

    if hasattr(schema, "_is_runtime_protocol") and schema._is_runtime_protocol:  # type: ignore[union-attr]
        return schema.__name__  # type: ignore[union-attr]

    return None


class NodeStoreBase:
    """Internal Store for InfrahubNode objects.

    Often while creating a lot of new objects,
    we need to save them in order to reuse them later to associate them with another node for example.
    """

    def __init__(self) -> None:
        self._store: dict[str, dict] = defaultdict(dict)
        self._store_by_hfid: dict[str, Any] = defaultdict(dict)

    def _set(self, node: InfrahubNode | InfrahubNodeSync | SchemaType, key: str | None = None) -> None:
        hfid = node.get_human_friendly_id_as_string(include_kind=True)

        if not key and not hfid:
            raise ValueError("Cannot store node without human friendly ID or key.")

        if key:
            node_kind = node._schema.kind
            self._store[node_kind][key] = node

        if hfid:
            self._store_by_hfid[hfid] = node

    def _get(self, key: str, kind: str | type[SchemaType] | None = None, raise_when_missing: bool = True):  # type: ignore[no-untyped-def]
        kind_name = get_schema_name(schema=kind)
        if kind_name and kind_name not in self._store and key not in self._store[kind_name]:  # type: ignore[attr-defined]
            if not raise_when_missing:
                return None
            raise NodeNotFoundError(
                node_type=kind_name,
                identifier={"key": [key]},
                message="Unable to find the node in the Store",
            )

        if kind_name and kind_name in self._store and key in self._store[kind_name]:  # type: ignore[attr-defined]
            return self._store[kind_name][key]  # type: ignore[attr-defined]

        for item in self._store.values():  # type: ignore[attr-defined]
            if key in item:
                return item[key]

        if not raise_when_missing:
            return None
        raise NodeNotFoundError(
            node_type="n/a",
            identifier={"key": [key]},
            message=f"Unable to find the node {key!r} in the store",
        )

    def _get_by_hfid(self, key: str, raise_when_missing: bool = True):  # type: ignore[no-untyped-def]
        try:
            return self._store_by_hfid[key]
        except KeyError as exc:
            if raise_when_missing:
                raise NodeNotFoundError(
                    node_type="n/a",
                    identifier={"key": [key]},
                    message=f"Unable to find the node {key!r} in the store",
                ) from exc
        return None


class NodeStore(NodeStoreBase):
    @overload
    def get(self, key: str, kind: type[SchemaType], raise_when_missing: Literal[True] = True) -> SchemaType: ...

    @overload
    def get(
        self, key: str, kind: type[SchemaType], raise_when_missing: Literal[False] = False
    ) -> SchemaType | None: ...

    @overload
    def get(self, key: str, kind: type[SchemaType], raise_when_missing: bool = ...) -> SchemaType: ...

    @overload
    def get(
        self, key: str, kind: str | None = ..., raise_when_missing: Literal[False] = False
    ) -> InfrahubNode | None: ...

    @overload
    def get(self, key: str, kind: str | None = ..., raise_when_missing: Literal[True] = True) -> InfrahubNode: ...

    @overload
    def get(self, key: str, kind: str | None = ..., raise_when_missing: bool = ...) -> InfrahubNode: ...

    def get(
        self, key: str, kind: str | type[SchemaType] | None = None, raise_when_missing: bool = True
    ) -> InfrahubNode | SchemaType | None:
        return self._get(key=key, kind=kind, raise_when_missing=raise_when_missing)

    @overload
    def get_by_hfid(self, key: str, raise_when_missing: Literal[True] = True) -> InfrahubNode: ...

    @overload
    def get_by_hfid(self, key: str, raise_when_missing: Literal[False] = False) -> InfrahubNode | None: ...

    def get_by_hfid(self, key: str, raise_when_missing: bool = True) -> InfrahubNode | None:
        return self._get_by_hfid(key=key, raise_when_missing=raise_when_missing)

    def set(self, node: Any, key: str | None = None) -> None:
        return self._set(node=node, key=key)


class NodeStoreSync(NodeStoreBase):
    @overload
    def get(self, key: str, kind: str | None = None, raise_when_missing: Literal[True] = True) -> InfrahubNodeSync: ...

    @overload
    def get(
        self, key: str, kind: str | None = None, raise_when_missing: Literal[False] = False
    ) -> InfrahubNodeSync | None: ...

    def get(self, key: str, kind: str | None = None, raise_when_missing: bool = True) -> InfrahubNodeSync | None:
        return self._get(key=key, kind=kind, raise_when_missing=raise_when_missing)

    @overload
    def get_by_hfid(self, key: str, raise_when_missing: Literal[True] = True) -> InfrahubNodeSync: ...

    @overload
    def get_by_hfid(self, key: str, raise_when_missing: Literal[False] = False) -> InfrahubNodeSync | None: ...

    def get_by_hfid(self, key: str, raise_when_missing: bool = True) -> InfrahubNodeSync | None:
        return self._get_by_hfid(key=key, raise_when_missing=raise_when_missing)

    def set(self, node: InfrahubNodeSync, key: str | None = None) -> None:
        return self._set(node=node, key=key)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.account import GlobalPermission
from infrahub.core.constants import GLOBAL_BRANCH_NAME, GlobalPermissions, InfrahubKind, PermissionDecision
from infrahub.core.schema.node_schema import NodeSchema
from infrahub.permissions.constants import BranchRelativePermissionDecision, PermissionDecisionFlag

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.permissions.manager import PermissionManager
    from infrahub.permissions.types import KindPermissions


__all__ = ["report_schema_permissions"]


def get_permission_report(  # noqa: PLR0911
    permission_manager: PermissionManager,
    branch: Branch,
    node: MainSchemaTypes,
    action: str,
    global_permission_report: dict[GlobalPermissions, bool],
) -> BranchRelativePermissionDecision:
    if global_permission_report[GlobalPermissions.SUPER_ADMIN]:
        return BranchRelativePermissionDecision.ALLOW

    if action != "view":
        if node.kind in (InfrahubKind.ACCOUNTGROUP, InfrahubKind.ACCOUNTROLE, InfrahubKind.GENERICACCOUNT) or (
            isinstance(node, NodeSchema) and InfrahubKind.GENERICACCOUNT in node.inherit_from
        ):
            return (
                BranchRelativePermissionDecision.ALLOW
                if global_permission_report[GlobalPermissions.MANAGE_ACCOUNTS]
                else BranchRelativePermissionDecision.DENY
            )
        if node.kind in (InfrahubKind.BASEPERMISSION, InfrahubKind.GLOBALPERMISSION, InfrahubKind.OBJECTPERMISSION) or (
            isinstance(node, NodeSchema) and InfrahubKind.BASEPERMISSION in node.inherit_from
        ):
            return (
                BranchRelativePermissionDecision.ALLOW
                if global_permission_report[GlobalPermissions.MANAGE_PERMISSIONS]
                else BranchRelativePermissionDecision.DENY
            )
        if node.kind in (InfrahubKind.GENERICREPOSITORY, InfrahubKind.REPOSITORY, InfrahubKind.READONLYREPOSITORY) or (
            isinstance(node, NodeSchema) and InfrahubKind.GENERICREPOSITORY in node.inherit_from
        ):
            return (
                BranchRelativePermissionDecision.ALLOW
                if global_permission_report[GlobalPermissions.MANAGE_REPOSITORIES]
                else BranchRelativePermissionDecision.DENY
            )

    is_default_branch = branch.name in (GLOBAL_BRANCH_NAME, registry.default_branch)
    decision = permission_manager.report_object_permission(namespace=node.namespace, name=node.name, action=action)

    if (
        decision == PermissionDecisionFlag.ALLOW_ALL
        or (decision & PermissionDecisionFlag.ALLOW_DEFAULT and is_default_branch)
        or (decision & PermissionDecisionFlag.ALLOW_OTHER and not is_default_branch)
    ):
        return BranchRelativePermissionDecision.ALLOW
    if decision & PermissionDecisionFlag.ALLOW_DEFAULT:
        return BranchRelativePermissionDecision.ALLOW_DEFAULT
    if decision & PermissionDecisionFlag.ALLOW_OTHER:
        return BranchRelativePermissionDecision.ALLOW_OTHER

    return BranchRelativePermissionDecision.DENY


async def report_schema_permissions(
    branch: Branch, permission_manager: PermissionManager, schemas: list[MainSchemaTypes]
) -> list[KindPermissions]:
    global_permission_report: dict[GlobalPermissions, bool] = {}
    for perm in GlobalPermissions:
        global_permission_report[perm] = permission_manager.resolve_global_permission(
            permission_to_check=GlobalPermission(action=perm.value, decision=PermissionDecision.ALLOW_ALL.value)
        )

    permission_objects: list[KindPermissions] = []
    for node in schemas:
        permission_objects.append(
            {
                "kind": node.kind,
                "create": get_permission_report(
                    permission_manager=permission_manager,
                    branch=branch,
                    node=node,
                    action="create",
                    global_permission_report=global_permission_report,
                ),
                "delete": get_permission_report(
                    permission_manager=permission_manager,
                    branch=branch,
                    node=node,
                    action="delete",
                    global_permission_report=global_permission_report,
                ),
                "update": get_permission_report(
                    permission_manager=permission_manager,
                    branch=branch,
                    node=node,
                    action="update",
                    global_permission_report=global_permission_report,
                ),
                "view": get_permission_report(
                    permission_manager=permission_manager,
                    branch=branch,
                    node=node,
                    action="view",
                    global_permission_report=global_permission_report,
                ),
            }
        )

    return permission_objects



from __future__ import annotations

from datetime import timedelta

from prefect import flow
from prefect.automations import AutomationCore
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.actions import RunDeployment
from prefect.events.schemas.automations import EventTrigger, Posture
from prefect.logging import get_run_logger

from infrahub.workflows.catalogue import COMPUTED_ATTRIBUTE_SETUP, COMPUTED_ATTRIBUTE_SETUP_PYTHON

from .constants import AUTOMATION_NAME


@flow(name="schema-updated-setup", flow_run_name="Setup schema updated event in task-manager")
async def schema_updated_setup() -> None:
    log = get_run_logger()

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(
                    name=DeploymentFilterName(
                        any_=[COMPUTED_ATTRIBUTE_SETUP.name, COMPUTED_ATTRIBUTE_SETUP_PYTHON.name]
                    )
                )
            )
        }
        if COMPUTED_ATTRIBUTE_SETUP.name not in deployments:
            raise ValueError("Unable to find the deployment for PROCESS_COMPUTED_MACRO")

        deployment_id_computed_attribute_setup = deployments[COMPUTED_ATTRIBUTE_SETUP.name].id
        deployment_id_computed_attribute_setup_python = deployments[COMPUTED_ATTRIBUTE_SETUP_PYTHON.name].id

        schema_update_automation = await client.find_automation(id_or_name=AUTOMATION_NAME)

        automation = AutomationCore(
            name=AUTOMATION_NAME,
            description="Trigger actions on schema update event",
            enabled=True,
            trigger=EventTrigger(
                posture=Posture.Reactive,
                expect={"infrahub.schema.update"},
                within=timedelta(0),
                threshold=1,
            ),
            actions=[
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_computed_attribute_setup,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                    },
                    job_variables={},
                ),
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_computed_attribute_setup_python,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                    },
                    job_variables={},
                ),
            ],
        )

        if schema_update_automation:
            await client.update_automation(automation_id=schema_update_automation.id, automation=automation)
            log.info(f"{AUTOMATION_NAME} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{AUTOMATION_NAME} Created")



from typing import Any

from pydantic import BaseModel, Field

from infrahub.message_bus import InfrahubMessage, Meta

from .constants import EVENT_NAMESPACE


class EventMeta(BaseModel):
    request_id: str = ""
    account_id: str = ""
    initiator_id: str | None = Field(
        default=None, description="The worker identity of the initial sender of this message"
    )


class InfrahubEvent(BaseModel):
    meta: EventMeta | None = None

    def get_event_namespace(self) -> str:
        return EVENT_NAMESPACE

    def get_name(self) -> str:
        return f"{self.get_event_namespace()}.unknown"

    def get_resource(self) -> dict[str, str]:
        raise NotImplementedError

    def get_messages(self) -> list[InfrahubMessage]:
        raise NotImplementedError

    def get_related(self) -> list[dict[str, str]]:
        related: list[dict[str, str]] = []

        if not self.meta:
            return related

        if self.meta.account_id:
            related.append(
                {
                    "prefect.resource.id": f"infrahub.account.{self.meta.account_id}",
                    "prefect.resource.role": "account",
                }
            )

        if self.meta.request_id:
            related.append(
                {
                    "prefect.resource.id": f"infrahub.request.{self.meta.request_id}",
                    "prefect.resource.role": "request",
                }
            )

        if self.meta.initiator_id:
            related.append(
                {
                    "prefect.resource.id": f"infrahub.source.{self.meta.initiator_id}",
                    "prefect.resource.role": "event_source",
                }
            )

        return related

    def get_payload(self) -> dict[str, Any]:
        return {}

    def get_message_meta(self) -> Meta:
        meta = Meta()
        if not self.meta:
            return meta

        if self.meta.initiator_id:
            meta.initiator_id = self.meta.initiator_id
        if self.meta.request_id:
            meta.initiator_id = self.meta.request_id

        return meta


class InfrahubBranchEvent(InfrahubEvent):  # pylint: disable=abstract-method
    branch: str = Field(..., description="The branch on which the event happend")

    def get_related(self) -> list[dict[str, str]]:
        related = super().get_related()
        related.append(
            {
                "prefect.resource.id": "infrahub.branch",
                "prefect.resource.name": self.branch,
                "prefect.resource.role": "branch",
            }
        )
        return related



from typing import Any, Optional

from graphql import GraphQLSchema, OperationType
from infrahub_sdk.analyzer import GraphQLQueryAnalyzer
from infrahub_sdk.utils import extract_fields

from infrahub.core.branch import Branch
from infrahub.graphql.utils import extract_schema_models


class InfrahubGraphQLQueryAnalyzer(GraphQLQueryAnalyzer):
    def __init__(
        self,
        query: str,
        query_variables: Optional[dict[str, Any]] = None,
        schema: Optional[GraphQLSchema] = None,
        operation_name: Optional[str] = None,
        branch: Optional[Branch] = None,
    ) -> None:
        self.branch: Optional[Branch] = branch
        self.operation_name: Optional[str] = operation_name
        self.query_variables: dict[str, Any] = query_variables or {}
        super().__init__(query=query, schema=schema)

    @property
    def operation_names(self) -> list[str]:
        return [operation.name for operation in self.operations if operation.name is not None]

    async def get_models_in_use(self, types: dict[str, Any]) -> set[str]:
        """List of Infrahub models that are referenced in the query."""
        graphql_types = set()
        models = set()

        if not (self.schema and self.branch):
            raise ValueError("Schema and Branch must be provided to extract the models in use.")

        for definition in self.document.definitions:
            fields = await extract_fields(definition.selection_set)

            operation = getattr(definition, "operation", None)
            if operation == OperationType.QUERY:
                schema = self.schema.query_type
            elif operation == OperationType.MUTATION:
                schema = self.schema.mutation_type
            else:
                # Subscription not supported right now
                continue

            graphql_types.update(await extract_schema_models(fields=fields, schema=schema, root_schema=self.schema))

        for graphql_type_name in graphql_types:
            try:
                graphql_type = types.get(graphql_type_name)
                if not hasattr(graphql_type, "_meta") or not hasattr(graphql_type._meta, "schema"):  # type: ignore[union-attr]
                    continue
                models.add(graphql_type._meta.schema.kind)  # type: ignore[union-attr]
            except ValueError:
                continue

        return models



from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.diff.query_parser import DiffQueryParser
from infrahub.core.query.diff import DiffAllPathsQuery
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger

from .model.path import CalculatedDiffs, NodeFieldSpecifier

log = get_logger()


class DiffCalculator:
    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db

    async def calculate_diff(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: Timestamp,
        to_time: Timestamp,
        include_unchanged: bool = True,
        previous_node_specifiers: set[NodeFieldSpecifier] | None = None,
    ) -> CalculatedDiffs:
        if diff_branch.name == registry.default_branch:
            diff_branch_from_time = from_time
        else:
            diff_branch_from_time = Timestamp(diff_branch.get_branched_from())
        diff_parser = DiffQueryParser(
            base_branch=base_branch,
            diff_branch=diff_branch,
            schema_manager=registry.schema,
            from_time=from_time,
            to_time=to_time,
            previous_node_field_specifiers=previous_node_specifiers,
        )
        branch_diff_query = await DiffAllPathsQuery.init(
            db=self.db,
            branch=diff_branch,
            base_branch=base_branch,
            diff_branch_from_time=diff_branch_from_time,
            diff_from=from_time,
            diff_to=to_time,
        )
        log.info("Beginning diff calculation query for branch")
        await branch_diff_query.execute(db=self.db)
        log.info("Diff calculation query for branch complete")
        log.info("Reading results of query for branch")
        for query_result in branch_diff_query.get_results():
            diff_parser.read_result(query_result=query_result)
        log.info("Results of query for branch read")

        if base_branch.name != diff_branch.name:
            new_node_field_specifiers = diff_parser.get_new_node_field_specifiers()
            current_node_field_specifiers = diff_parser.get_current_node_field_specifiers()

            base_diff_query = await DiffAllPathsQuery.init(
                db=self.db,
                branch=base_branch,
                base_branch=base_branch,
                diff_branch_from_time=diff_branch_from_time,
                diff_from=from_time,
                diff_to=to_time,
                current_node_field_specifiers=[
                    (nfs.node_uuid, nfs.field_name) for nfs in current_node_field_specifiers
                ],
                new_node_field_specifiers=[(nfs.node_uuid, nfs.field_name) for nfs in new_node_field_specifiers],
            )

            log.info("Beginning diff calculation query for base")
            await base_diff_query.execute(db=self.db)
            log.info("Diff calculation query for base complete")
            log.info("Reading results of query for base")
            for query_result in base_diff_query.get_results():
                diff_parser.read_result(query_result=query_result)
            log.info("Results of query for branch read")
        log.info("Parsing calculated diff")
        diff_parser.parse(include_unchanged=include_unchanged)
        log.info("Calculated diff parsed")
        return CalculatedDiffs(
            base_branch_name=base_branch.name,
            diff_branch_name=diff_branch.name,
            base_branch_diff=diff_parser.get_diff_root_for_branch(branch=base_branch.name),
            diff_branch_diff=diff_parser.get_diff_root_for_branch(branch=diff_branch.name),
        )



from __future__ import annotations

import asyncio
import os
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from infrahub_sdk.repository import GitRepoManager

from .exceptions import UninitializedError

if TYPE_CHECKING:
    from . import InfrahubClient

INFRAHUB_TRANSFORM_VARIABLE_TO_IMPORT = "INFRAHUB_TRANSFORMS"


class InfrahubTransform:
    name: str | None = None
    query: str
    timeout: int = 10

    def __init__(
        self,
        branch: str = "",
        root_directory: str = "",
        server_url: str = "",
        client: InfrahubClient | None = None,
    ):
        self.git: GitRepoManager

        self.branch = branch
        self.server_url = server_url or os.environ.get("INFRAHUB_URL", "http://127.0.0.1:8000")
        self.root_directory = root_directory or os.getcwd()

        self._client = client

        if not self.name:
            self.name = self.__class__.__name__

        if not self.query:
            raise ValueError("A query must be provided")

    @property
    def client(self) -> InfrahubClient:
        if self._client:
            return self._client

        raise UninitializedError("The client has not been initialized")

    @property
    def branch_name(self) -> str:
        """Return the name of the current git branch."""

        if self.branch:
            return self.branch

        if not hasattr(self, "git") or not self.git:
            self.git = GitRepoManager(self.root_directory)

        self.branch = str(self.git.active_branch)

        return self.branch

    @abstractmethod
    def transform(self, data: dict) -> Any:
        pass

    async def collect_data(self) -> dict:
        """Query the result of the GraphQL Query defined in self.query and return the result"""

        return await self.client.query_gql_query(name=self.query, branch_name=self.branch_name)

    async def run(self, data: dict | None = None) -> Any:
        """Execute the transformation after collecting the data from the GraphQL query.

        The result of the check is determined based on the presence or not of ERROR log messages.

        Args:
            data: The data on which to run the transform. Data will be queried from the API if not provided

        Returns: Transformed data
        """

        if not data:
            data = await self.collect_data()

        unpacked = data.get("data") or data

        if asyncio.iscoroutinefunction(self.transform):
            return await self.transform(data=unpacked)

        return self.transform(data=unpacked)



from __future__ import annotations

from pydantic import Field

from infrahub.core.constants import InfrahubKind
from infrahub.core.schema.basenode_schema import BaseNodeSchema


class ProfileSchema(BaseNodeSchema):
    inherit_from: list[str] = Field(
        default_factory=list,
        description="List of Generic Kind that this profile is inheriting from",
    )

    @property
    def is_node_schema(self) -> bool:
        return False

    @property
    def is_generic_schema(self) -> bool:
        return False

    @property
    def is_profile_schema(self) -> bool:
        return True

    def get_labels(self) -> list[str]:
        """Return the labels for this object, composed of the kind
        and the list of Generic this object is inheriting from."""

        labels: list[str] = [self.kind] + self.inherit_from
        if self.namespace not in ["Schema", "Internal"] and InfrahubKind.GENERICGROUP not in self.inherit_from:
            labels.append(InfrahubKind.PROFILE)
        return labels



from typing import TYPE_CHECKING, Any, Optional

from graphene import Boolean, Field, InputObjectType, Mutation, String
from graphql import GraphQLResolveInfo

from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch
from infrahub.core.constants import (
    CheckType,
    GlobalPermissions,
    InfrahubKind,
    PermissionDecision,
)
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import BranchNotFoundError, PermissionDeniedError, ValidationError
from infrahub.graphql.mutations.main import InfrahubMutationMixin
from infrahub.graphql.types.enums import CheckType as GraphQLCheckType
from infrahub.message_bus import messages
from infrahub.proposed_change.constants import ProposedChangeState
from infrahub.workflows.catalogue import PROPOSED_CHANGE_MERGE

from ..types.task import TaskInfo
from .main import InfrahubMutationOptions

if TYPE_CHECKING:
    from ..initialization import GraphqlContext


class InfrahubProposedChangeMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(cls, schema: NodeSchema = None, _meta=None, **options):  # pylint: disable=arguments-differ
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    @retry_db_transaction(name="proposed_change_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ):
        context: GraphqlContext = info.context
        db: InfrahubDatabase = info.context.db

        async with db.start_transaction() as dbt:
            proposed_change, result = await super().mutate_create(info=info, data=data, branch=branch, database=dbt)
            destination_branch = proposed_change.destination_branch.value
            source_branch = await _get_source_branch(db=dbt, name=proposed_change.source_branch.value)
            if destination_branch == source_branch.name:
                raise ValidationError(input_value="The source and destination branch can't be the same")
            if destination_branch != "main":
                raise ValidationError(
                    input_value="Currently only the 'main' branch is supported as a destination for a proposed change"
                )

        if context.service:
            message_list = [
                messages.RequestProposedChangePipeline(
                    proposed_change=proposed_change.id,
                    source_branch=source_branch.name,
                    source_branch_sync_with_git=source_branch.sync_with_git,
                    destination_branch=destination_branch,
                ),
            ]

            for message in message_list:
                await context.service.send(message=message)

        return proposed_change, result

    @classmethod
    async def mutate_update(  # pylint: disable=too-many-branches
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ):
        context: GraphqlContext = info.context

        obj = await NodeManager.get_one_by_id_or_default_filter(
            db=context.db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        state = ProposedChangeState(obj.state.value.value)
        state.validate_editability()

        updated_state = None
        if state_update := data.get("state", {}).get("value"):
            updated_state = ProposedChangeState(state_update)
            state.validate_state_transition(updated_state)

        # Check before starting a transaction, stopping in the middle of the transaction seems to break with memgraph
        if updated_state == ProposedChangeState.MERGED and context.account_session:
            try:
                context.active_permissions.raise_for_permission(
                    permission=GlobalPermission(
                        action=GlobalPermissions.MERGE_PROPOSED_CHANGE.value,
                        decision=PermissionDecision.ALLOW_ALL.value,
                    )
                )
            except PermissionDeniedError as exc:
                raise ValidationError(str(exc)) from exc

        if updated_state == ProposedChangeState.MERGED:
            data["state"]["value"] = ProposedChangeState.MERGING.value

        proposed_change, result = await super().mutate_update(
            info=info, data=data, branch=branch, database=context.db, node=obj
        )

        if updated_state == ProposedChangeState.MERGED:
            await context.service.workflow.execute_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )

        return proposed_change, result


class ProposedChangeRequestRunCheckInput(InputObjectType):
    id = String(required=True)
    check_type = GraphQLCheckType(required=False)


class ProposedChangeRequestRunCheck(Mutation):
    class Arguments:
        data = ProposedChangeRequestRunCheckInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # pylint: disable=unused-argument
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> dict[str, bool]:
        context: GraphqlContext = info.context

        check_type = data.get("check_type") or CheckType.ALL

        identifier = data.get("id", "")
        proposed_change = await NodeManager.get_one_by_id_or_default_filter(
            id=identifier, kind=InfrahubKind.PROPOSEDCHANGE, db=context.db
        )
        state = ProposedChangeState(proposed_change.state.value.value)
        state.validate_state_check_run()

        destination_branch = proposed_change.destination_branch.value
        source_branch = await _get_source_branch(db=context.db, name=proposed_change.source_branch.value)

        message = messages.RequestProposedChangePipeline(
            proposed_change=proposed_change.id,
            source_branch=source_branch.name,
            source_branch_sync_with_git=source_branch.sync_with_git,
            destination_branch=destination_branch,
            check_type=check_type,
        )
        if context.service:
            await context.service.send(message=message)

        return {"ok": True}


class ProposedChangeMergeInput(InputObjectType):
    id = String(required=True)


class ProposedChangeMerge(Mutation):
    class Arguments:
        data = ProposedChangeMergeInput(required=True)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # pylint: disable=unused-argument
        info: GraphQLResolveInfo,
        data: dict[str, Any],
        wait_until_completion: bool = True,
    ) -> dict[str, bool]:
        context: GraphqlContext = info.context
        task: dict | None = None

        identifier = data.get("id", "")
        proposed_change = await NodeManager.get_one(
            id=identifier, kind=InfrahubKind.PROPOSEDCHANGE, db=context.db, raise_on_error=True
        )
        state = ProposedChangeState(proposed_change.state.value.value)
        if state != ProposedChangeState.OPEN:
            raise ValidationError("Only proposed change in OPEN state can be merged")

        async with context.db.start_session() as db:
            proposed_change.state.value = ProposedChangeState.MERGING.value
            proposed_change.save(db=db)

        if wait_until_completion:
            await context.service.workflow.execute_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
        else:
            workflow = await context.service.workflow.submit_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
            task = {"id": workflow.id}

        return cls(ok=True, task=task)


async def _get_source_branch(db: InfrahubDatabase, name: str) -> Branch:
    try:
        return await Branch.get_by_name(name=name, db=db)
    except BranchNotFoundError:
        raise ValidationError(
            input_value="The specified source branch for this proposed change was not found."
        ) from None



from abc import ABC, abstractmethod

from infrahub.core.path import GroupedDataPaths

from .model import SchemaConstraintValidatorRequest


class ConstraintCheckerInterface(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def supports(self, request: SchemaConstraintValidatorRequest) -> bool: ...

    @abstractmethod
    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Migration003Query01(Query):
    name = "migration_003_01"
    type: QueryType = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        at = Timestamp()
        filters, params = at.get_query_filter_path()

        self.params.update(params)

        # ruff: noqa: E501
        query = """
        MATCH path = (av2:AttributeValue)-[:HAS_VALUE]-(:Attribute {name: "optional"})-[:HAS_ATTRIBUTE]-(n:SchemaRelationship)-[:HAS_ATTRIBUTE]-(:Attribute {name: "kind"})-[:HAS_VALUE]-(av1:AttributeValue)
        WHERE av1.value = "Parent" AND av2.value = true AND all(r IN relationships(path) WHERE ( %(filters)s ))
        CALL {
            WITH n
            MATCH path22 = (av22:AttributeValue)-[r22:HAS_VALUE]-(a2:Attribute {name: "optional"})-[:HAS_ATTRIBUTE]-(n:SchemaRelationship)-[:HAS_ATTRIBUTE]-(:Attribute {name:"kind"})-[:HAS_VALUE]-(av11:AttributeValue)
            WHERE av11.value = "Parent" AND av22.value = true AND all(r IN relationships(path22) WHERE ( %(filters)s ))
            RETURN av22 as av2_sub, r22, a2, path22 as path2
            ORDER BY r22.branch_level DESC, r22.from DESC
            LIMIT 1
        }
        WITH av2_sub as av2, r22, a2, path2
        WHERE all(r IN relationships(path2) WHERE r.status = "active")
        MERGE (new_value: AttributeValue { value: false, is_default: false })
        CREATE (a2)-[:HAS_VALUE { branch: r22.branch, branch_level: r22.branch_level, status: "active", from: $at } ]->(new_value)
        SET r22.to = $at
        """ % {"filters": filters}
        self.add_to_query(query)
        self.return_labels = ["av2"]


class Migration003(GraphMigration):
    name: str = "003_relationship_parent_mandatory"
    queries: Sequence[type[Query]] = [Migration003Query01]
    minimum_version: int = 2

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        return result



from __future__ import annotations

from typing import TYPE_CHECKING

from .registry import refresh_branches

if TYPE_CHECKING:
    from infrahub.services import InfrahubServices


async def trigger_branch_refresh(service: InfrahubServices) -> None:
    service.log.debug("Running branch refresh task")
    async with service.database.start_session() as db:
        await refresh_branches(db=db)

    await service.component.refresh_schema_hash()



from typing import TYPE_CHECKING

from graphene import Boolean, DateTime, InputObjectType, Mutation, String
from graphql import GraphQLResolveInfo

from infrahub.core import registry
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.database import retry_db_transaction
from infrahub.dependencies.registry import get_component_registry
from infrahub.workflows.catalogue import DIFF_UPDATE

if TYPE_CHECKING:
    from ..initialization import GraphqlContext


class DiffUpdateInput(InputObjectType):
    branch = String(required=True)
    name = String(required=False)
    from_time = DateTime(required=False)
    to_time = DateTime(required=False)
    wait_for_completion = Boolean(required=False)


class DiffUpdateMutation(Mutation):
    class Arguments:
        data = DiffUpdateInput(required=True)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="diff_update")
    async def mutate(
        cls,
        root: dict,  # pylint: disable=unused-argument
        info: GraphQLResolveInfo,
        data: DiffUpdateInput,
    ) -> dict[str, bool]:
        context: GraphqlContext = info.context

        from_timestamp_str = DateTime.serialize(data.from_time) if data.from_time else None
        to_timestamp_str = DateTime.serialize(data.to_time) if data.to_time else None
        if data.wait_for_completion is True:
            component_registry = get_component_registry()
            base_branch = await registry.get_branch(db=context.db, branch=registry.default_branch)
            diff_branch = await registry.get_branch(db=context.db, branch=data.branch)

            diff_coordinator = await component_registry.get_component(
                DiffCoordinator, db=context.db, branch=diff_branch
            )
            await diff_coordinator.run_update(
                base_branch=base_branch,
                diff_branch=diff_branch,
                from_time=from_timestamp_str,
                to_time=to_timestamp_str,
                name=data.name,
            )

            return {"ok": True}

        model = RequestDiffUpdate(
            branch_name=str(data.branch),
            name=data.name,
            from_time=from_timestamp_str,
            to_time=to_timestamp_str,
        )
        if context.service:
            await context.service.workflow.submit_workflow(workflow=DIFF_UPDATE, parameters={"model": model})

        return {"ok": True}



from typing import Awaitable, Callable, Optional

from infrahub_sdk import InfrahubClient

from infrahub.components import ComponentType
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import InitializationError
from infrahub.log import get_logger
from infrahub.message_bus import InfrahubMessage, InfrahubResponse
from infrahub.message_bus.messages import ROUTING_KEY_MAP
from infrahub.message_bus.types import MessageTTL

from .adapters.cache import InfrahubCache
from .adapters.event import InfrahubEventService
from .adapters.http import InfrahubHTTP
from .adapters.http.httpx import HttpxAdapter
from .adapters.message_bus import InfrahubMessageBus
from .adapters.workflow import InfrahubWorkflow
from .component import InfrahubComponent
from .protocols import InfrahubLogger
from .scheduler import InfrahubScheduler


class InfrahubServices:
    def __init__(
        self,
        cache: Optional[InfrahubCache] = None,
        client: Optional[InfrahubClient] = None,
        database: Optional[InfrahubDatabase] = None,
        message_bus: Optional[InfrahubMessageBus] = None,
        http: InfrahubHTTP | None = None,
        workflow: Optional[InfrahubWorkflow] = None,
        event: InfrahubEventService | None = None,
        log: Optional[InfrahubLogger] = None,
        component_type: Optional[ComponentType] = None,
    ) -> None:
        self.cache = cache or InfrahubCache()
        self._client = client
        self._database = database
        self.message_bus = message_bus or InfrahubMessageBus()
        self.workflow = workflow or InfrahubWorkflow()
        self.event = event or InfrahubEventService()
        self.log = log or get_logger()
        self.component_type = component_type or ComponentType.NONE
        self.http = http or HttpxAdapter()
        self.scheduler = InfrahubScheduler()
        self.component = InfrahubComponent()

    @property
    def client(self) -> InfrahubClient:
        if not self._client:
            raise InitializationError("Service is not initialized with a client")

        return self._client

    def set_client(self, client: InfrahubClient | None) -> None:
        self._client = client

    @property
    def database(self) -> InfrahubDatabase:
        if not self._database:
            raise InitializationError("Service is not initialized with a database")

        return self._database

    async def initialize(self) -> None:
        """Initialize the Services"""
        await self.message_bus.initialize(service=self)
        await self.cache.initialize(service=self)
        await self.http.initialize(service=self)
        await self.component.initialize(service=self)
        await self.scheduler.initialize(service=self)
        await self.workflow.initialize(service=self)
        await self.event.initialize(service=self)

    async def shutdown(self) -> None:
        """Initialize the Services"""
        await self.scheduler.shutdown()
        await self.message_bus.shutdown()

    async def send(self, message: InfrahubMessage, delay: Optional[MessageTTL] = None, is_retry: bool = False) -> None:
        routing_key = ROUTING_KEY_MAP.get(type(message))
        if not routing_key:
            raise ValueError("Unable to determine routing key")
        await self.message_bus.publish(message, routing_key=routing_key, delay=delay, is_retry=is_retry)

    async def reply(self, message: InfrahubResponse, initiator: InfrahubMessage) -> None:
        if initiator.meta:
            message.meta.correlation_id = initiator.meta.correlation_id
            routing_key = initiator.meta.reply_to or ""
            await self.message_bus.reply(message, routing_key=routing_key)


class ServiceManager:
    def __init__(self) -> None:
        self.service = InfrahubServices()
        self.send = self.service.send

    def prepare(self, service: InfrahubServices) -> None:
        self.service = service
        self.send = self.service.send


ServiceFunction = Callable[[InfrahubServices], Awaitable[None]]


services = ServiceManager()



from infrahub.core.validators.node.relationship import NodeRelationshipAddChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaNodeRelationshipAddConstraintDependency(DependencyBuilder[NodeRelationshipAddChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeRelationshipAddChecker:
        return NodeRelationshipAddChecker(db=context.db, branch=context.branch)



from ..model.path import CalculatedDiffs, EnrichedDiffs
from .interface import DiffEnricherInterface


class AggregatedDiffEnricher:
    def __init__(self, enrichers: list[DiffEnricherInterface]) -> None:
        self.enrichers = enrichers

    async def enrich(self, calculated_diffs: CalculatedDiffs) -> EnrichedDiffs:
        enriched_diffs = EnrichedDiffs.from_calculated_diffs(calculated_diffs=calculated_diffs)

        for enricher in self.enrichers:
            await enricher.enrich(enriched_diff_root=enriched_diffs.base_branch_diff, calculated_diffs=calculated_diffs)
            await enricher.enrich(enriched_diff_root=enriched_diffs.diff_branch_diff, calculated_diffs=calculated_diffs)

        return enriched_diffs



from __future__ import annotations

import asyncio
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from infrahub_sdk.protocols import CoreGeneratorDefinition, CoreProposedChange
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.client.schemas.objects import (
    State,  # noqa: TC002
)
from prefect.logging import get_run_logger
from prefect.states import Completed, Failed

from infrahub import config
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.tasks import merge_branch
from infrahub.core.constants import InfrahubKind, RepositoryInternalStatus, ValidatorConclusion
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.model.diff import DiffElementType, SchemaConflict
from infrahub.core.diff.model.path import NodeDiffFieldSummary
from infrahub.core.integrity.object_conflict.conflict_recorder import ObjectConflictValidatorRecorder
from infrahub.core.protocols import CoreDataCheck, CoreValidator
from infrahub.core.protocols import CoreProposedChange as InternalCoreProposedChange
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.generators.models import ProposedChangeGeneratorDefinition
from infrahub.git.repository import get_initialized_repo
from infrahub.log import get_logger
from infrahub.message_bus import InfrahubMessage, messages
from infrahub.message_bus.operations.requests.proposed_change import DefinitionSelect
from infrahub.proposed_change.constants import ProposedChangeState
from infrahub.proposed_change.models import (
    RequestProposedChangeDataIntegrity,
    RequestProposedChangeRepositoryChecks,
    RequestProposedChangeRunGenerators,
    RequestProposedChangeSchemaIntegrity,
    RequestProposedChangeUserTests,
)
from infrahub.pytest_plugin import InfrahubBackendPlugin
from infrahub.services import services
from infrahub.workflows.catalogue import COMPUTED_ATTRIBUTE_SETUP_PYTHON, REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS
from infrahub.workflows.utils import add_tags

if TYPE_CHECKING:
    from infrahub_sdk.node import InfrahubNode

    from infrahub.core.models import SchemaUpdateConstraintInfo
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.message_bus.types import ProposedChangeRepository


async def _proposed_change_transition_state(
    state: ProposedChangeState,
    proposed_change: InternalCoreProposedChange | None = None,
    proposed_change_id: str | None = None,
) -> None:
    service = services.service
    async with service.database.start_session() as db:
        if proposed_change is None and proposed_change_id:
            proposed_change = await registry.manager.get_one(
                db=db, id=proposed_change_id, kind=InternalCoreProposedChange, raise_on_error=True
            )
        if proposed_change:
            proposed_change.state.value = state.value  # type: ignore[misc]
            await proposed_change.save(db=db)


# async def proposed_change_transition_merged(flow: Flow, flow_run: FlowRun, state: State) -> None:
#     await _proposed_change_transition_state(
#         proposed_change_id=flow_run.parameters["proposed_change_id"], state=ProposedChangeState.MERGED
#     )


# async def proposed_change_transition_open(flow: Flow, flow_run: FlowRun, state: State) -> None:
#     await _proposed_change_transition_state(
#         proposed_change_id=flow_run.parameters["proposed_change_id"], state=ProposedChangeState.OPEN
#     )


@flow(
    name="proposed-change-merge",
    flow_run_name="Merge propose change: {proposed_change_name} ",
    description="Merge a given proposed change.",
    # TODO need to investigate why these function are not working as expected
    # on_completion=[proposed_change_transition_merged],  # type: ignore
    # on_failure=[proposed_change_transition_open],  # type: ignore
    # on_crashed=[proposed_change_transition_open],  # type: ignore
    # on_cancellation=[proposed_change_transition_open],  # type: ignore
)
async def merge_proposed_change(proposed_change_id: str, proposed_change_name: str) -> State:  # pylint: disable=unused-argument
    service = services.service
    log = get_run_logger()

    await add_tags(nodes=[proposed_change_id])

    async with service.database.start_session() as db:
        proposed_change = await registry.manager.get_one(
            db=db, id=proposed_change_id, kind=InternalCoreProposedChange, raise_on_error=True
        )

        log.info("Validating if all conditions are met to merge the proposed change")

        source_branch = await Branch.get_by_name(db=db, name=proposed_change.source_branch.value)
        validations = await proposed_change.validations.get_peers(db=db, peer_type=CoreValidator)
        for validation in validations.values():
            validator_kind = validation.get_kind()
            if (
                validator_kind != InfrahubKind.DATAVALIDATOR
                and validation.conclusion.value.value != ValidatorConclusion.SUCCESS.value
            ):
                # Ignoring Data integrity checks as they are handled again later
                await _proposed_change_transition_state(proposed_change=proposed_change, state=ProposedChangeState.OPEN)
                return Failed(message="Unable to merge proposed change containing failing checks")
            if validator_kind == InfrahubKind.DATAVALIDATOR:
                data_checks = await validation.checks.get_peers(db=db, peer_type=CoreDataCheck)
                for check in data_checks.values():
                    if check.conflicts.value and not check.keep_branch.value:
                        await _proposed_change_transition_state(
                            proposed_change=proposed_change, state=ProposedChangeState.OPEN
                        )
                        return Failed(
                            message="Data conflicts found on branch and missing decisions about what branch to keep"
                        )

        log.info("Proposed change is eligible to be merged")
        await merge_branch(branch=source_branch.name)

        log.info(f"Branch {source_branch.name} has been merged successfully")

        await _proposed_change_transition_state(proposed_change=proposed_change, state=ProposedChangeState.MERGED)
        await service.workflow.submit_workflow(workflow=COMPUTED_ATTRIBUTE_SETUP_PYTHON)
        return Completed(message="proposed change merged successfully")


@flow(
    name="proposed-changes-cancel-branch",
    flow_run_name="Cancel all proposed change associated with branch {branch_name}",
    description="Cancel all Proposed change associated with a branch.",
)
async def cancel_proposed_changes_branch(branch_name: str) -> None:
    service = services.service

    await add_tags(branches=[branch_name])

    proposed_changed_opened = await service.client.filters(
        kind=CoreProposedChange,
        include=["id", "source_branch"],
        state__value=ProposedChangeState.OPEN.value,
        source_branch__value=branch_name,
    )
    proposed_changed_closed = await service.client.filters(
        kind=CoreProposedChange,
        include=["id", "source_branch"],
        state__value=ProposedChangeState.CLOSED.value,
        source_branch__value=branch_name,
    )

    for proposed_change in proposed_changed_opened + proposed_changed_closed:
        await cancel_proposed_change(proposed_change=proposed_change)


@task(name="Cancel a propose change", description="Cancel a propose change", cache_policy=NONE)
async def cancel_proposed_change(proposed_change: CoreProposedChange) -> None:
    service = services.service

    await add_tags(nodes=[proposed_change.id])
    log = get_run_logger()

    log.info("Canceling proposed change as the source branch was deleted")
    proposed_change = await service.client.get(kind=CoreProposedChange, id=proposed_change.id)
    proposed_change.state.value = ProposedChangeState.CANCELED.value
    await proposed_change.save()


@flow(
    name="proposed-changed-data-integrity",
    flow_run_name="Triggers data integrity check",
)
async def run_proposed_change_data_integrity_check(model: RequestProposedChangeDataIntegrity) -> None:
    """Triggers a data integrity validation check on the provided proposed change to start."""
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    service = services.service
    async with service.database.start_transaction() as dbt:
        destination_branch = await registry.get_branch(db=dbt, branch=model.destination_branch)
        source_branch = await registry.get_branch(db=dbt, branch=model.source_branch)
        component_registry = get_component_registry()

        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=dbt, branch=source_branch)
        await diff_coordinator.update_branch_diff(base_branch=destination_branch, diff_branch=source_branch)


@flow(
    name="proposed-changed-run-generator",
    flow_run_name="Run generators",
)
async def run_generators(model: RequestProposedChangeRunGenerators) -> None:
    service = services.service
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change], db_change=True)

    generators = await service.client.filters(
        kind=CoreGeneratorDefinition,
        prefetch_relationships=True,
        populate_store=True,
        branch=model.source_branch,
    )
    generator_definitions = [
        ProposedChangeGeneratorDefinition(
            definition_id=generator.id,
            definition_name=generator.name.value,
            class_name=generator.class_name.value,
            file_path=generator.file_path.value,
            query_name=generator.query.peer.name.value,
            query_models=generator.query.peer.models.value,
            repository_id=generator.repository.peer.id,
            parameters=generator.parameters.value,
            group_id=generator.targets.peer.id,
            convert_query_response=generator.convert_query_response.value,
        )
        for generator in generators
    ]

    for generator_definition in generator_definitions:
        # Request generator definitions if the source branch that is managed in combination
        # to the Git repository containing modifications which could indicate changes to the transforms
        # in code
        # Alternatively if the queries used touches models that have been modified in the path
        # impacted artifact definitions will be included for consideration

        select = DefinitionSelect.NONE
        select = select.add_flag(
            current=select,
            flag=DefinitionSelect.FILE_CHANGES,
            condition=model.source_branch_sync_with_git and model.branch_diff.has_file_modifications,
        )

        for changed_model in model.branch_diff.modified_kinds(branch=model.source_branch):
            select = select.add_flag(
                current=select,
                flag=DefinitionSelect.MODIFIED_KINDS,
                condition=changed_model in generator_definition.query_models,
            )

        if select:
            msg = messages.RequestGeneratorDefinitionCheck(
                generator_definition=generator_definition,
                branch_diff=model.branch_diff,
                proposed_change=model.proposed_change,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                destination_branch=model.destination_branch,
            )
            msg.assign_meta(parent=model)
            await service.send(message=msg)

    next_messages: list[InfrahubMessage] = []
    if model.refresh_artifacts:
        next_messages.append(
            messages.RequestProposedChangeRefreshArtifacts(
                proposed_change=model.proposed_change,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                destination_branch=model.destination_branch,
                branch_diff=model.branch_diff,
            )
        )

    if model.do_repository_checks:
        model_proposed_change_repo_checks = RequestProposedChangeRepositoryChecks(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=model.branch_diff,
        )
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS, parameters={"model": model_proposed_change_repo_checks}
        )

    for next_msg in next_messages:
        next_msg.assign_meta(parent=model)
        await service.send(message=next_msg)


@flow(
    name="proposed-changed-schema-integrity",
    flow_run_name="Process schema integrity",
)
async def run_proposed_change_schema_integrity_check(
    model: RequestProposedChangeSchemaIntegrity,
) -> None:
    # For now, we retrieve the latest schema for each branch from the registry
    # In the future it would be good to generate the object SchemaUpdateValidationResult from message.branch_diff
    service = services.service
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    source_schema = registry.schema.get_schema_branch(name=model.source_branch).duplicate()
    dest_schema = registry.schema.get_schema_branch(name=model.destination_branch).duplicate()

    candidate_schema = dest_schema.duplicate()
    candidate_schema.update(schema=source_schema)
    schema_diff = dest_schema.diff(other=candidate_schema)
    validation_result = dest_schema.validate_update(other=candidate_schema, diff=schema_diff)

    constraints_from_data_diff = await _get_proposed_change_schema_integrity_constraints(
        model=model, schema=candidate_schema
    )
    constraints_from_schema_diff = validation_result.constraints
    constraints = set(constraints_from_data_diff + constraints_from_schema_diff)

    if not constraints:
        return

    # ----------------------------------------------------------
    # Validate if the new schema is valid with the content of the database
    # ----------------------------------------------------------
    source_branch = registry.get_branch_from_registry(branch=model.source_branch)
    responses = await schema_validate_migrations(
        message=SchemaValidateMigrationData(
            branch=source_branch, schema_branch=candidate_schema, constraints=list(constraints)
        )
    )

    # TODO we need to report a failure if an error happened during the execution of a validator
    conflicts: list[SchemaConflict] = []
    for response in responses:
        for violation in response.violations:
            conflicts.append(
                SchemaConflict(
                    name=response.schema_path.get_path(),
                    type=response.constraint_name,
                    kind=violation.node_kind,
                    id=violation.node_id,
                    path=response.schema_path.get_path(),
                    value=violation.message,
                    branch="placeholder",
                )
            )

    if not conflicts:
        return

    async with service.database.start_transaction() as db:
        object_conflict_validator_recorder = ObjectConflictValidatorRecorder(
            db=db,
            validator_kind=InfrahubKind.SCHEMAVALIDATOR,
            validator_label="Schema Integrity",
            check_schema_kind=InfrahubKind.SCHEMACHECK,
        )
        await object_conflict_validator_recorder.record_conflicts(
            proposed_change_id=model.proposed_change, conflicts=conflicts
        )


async def _get_proposed_change_schema_integrity_constraints(
    model: RequestProposedChangeSchemaIntegrity, schema: SchemaBranch
) -> list[SchemaUpdateConstraintInfo]:
    node_diff_field_summary_map: dict[str, NodeDiffFieldSummary] = {}
    for node_diff in model.branch_diff.diff_summary:
        node_kind = node_diff["kind"]
        if node_kind not in node_diff_field_summary_map:
            node_diff_field_summary_map[node_kind] = NodeDiffFieldSummary(kind=node_kind)
        field_summary = node_diff_field_summary_map[node_kind]
        for element in node_diff["elements"]:
            element_name = element["name"]
            element_type = element["element_type"]
            if element_type.lower() in (
                DiffElementType.RELATIONSHIP_MANY.value.lower(),
                DiffElementType.RELATIONSHIP_ONE.value.lower(),
            ):
                field_summary.relationship_names.add(element_name)
            elif element_type.lower() in (DiffElementType.ATTRIBUTE.value.lower(),):
                field_summary.attribute_names.add(element_name)

    determiner = ConstraintValidatorDeterminer(schema_branch=schema)
    return await determiner.get_constraints(node_diffs=list(node_diff_field_summary_map.values()))


@flow(
    name="proposed-changed-repository-checks",
    flow_run_name="Process user defined checks",
)
async def repository_checks(model: RequestProposedChangeRepositoryChecks) -> None:
    service = services.service
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    events: list[InfrahubMessage] = []
    for repository in model.branch_diff.repositories:
        if (
            model.source_branch_sync_with_git
            and not repository.read_only
            and repository.internal_status == RepositoryInternalStatus.ACTIVE.value
        ):
            events.append(
                messages.RequestRepositoryChecks(
                    proposed_change=model.proposed_change,
                    repository=repository.repository_id,
                    source_branch=model.source_branch,
                    target_branch=model.destination_branch,
                )
            )

        events.append(
            messages.RequestRepositoryUserChecks(
                proposed_change=model.proposed_change,
                repository_id=repository.repository_id,
                repository_name=repository.repository_name,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                target_branch=model.destination_branch,
                branch_diff=model.branch_diff,
            )
        )
    for event in events:
        event.assign_meta(parent=model)
        await service.send(message=event)


@flow(
    name="proposed-changed-user-tests",
    flow_run_name="Run unit tests in repositories",
)
async def run_proposed_change_user_tests(model: RequestProposedChangeUserTests) -> None:
    service = services.service
    log = get_run_logger()
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    def _execute(
        directory: Path, repository: ProposedChangeRepository, proposed_change: InfrahubNode
    ) -> int | pytest.ExitCode:
        config_file = str(directory / ".infrahub.yml")
        test_directory = directory / "tests"
        log = get_logger()

        if not test_directory.is_dir():
            log.debug(
                event="repository_tests_ignored",
                proposed_change=proposed_change,
                repository=repository.repository_name,
                message="tests directory not found",
            )
            return 1

        # Redirect stdout/stderr to avoid showing pytest lines in the git agent
        old_out = sys.stdout
        old_err = sys.stderr

        with Path(os.devnull).open(mode="w", encoding="utf-8") as devnull:
            sys.stdout = devnull
            sys.stderr = devnull

            exit_code = pytest.main(
                [
                    str(test_directory),
                    f"--infrahub-repo-config={config_file}",
                    f"--infrahub-address={config.SETTINGS.main.internal_address}",
                    "-qqqq",
                    "-s",
                ],
                plugins=[InfrahubBackendPlugin(service.client.config, repository.repository_id, proposed_change.id)],
            )

        # Restore stdout/stderr back to their orignal states
        sys.stdout = old_out
        sys.stderr = old_err

        return exit_code

    for repository in model.branch_diff.repositories:
        if model.source_branch_sync_with_git:
            repo = await get_initialized_repo(
                repository_id=repository.repository_id,
                name=repository.repository_name,
                service=service,
                repository_kind=repository.kind,
            )
            commit = repo.get_commit_value(proposed_change.source_branch.value)
            worktree_directory = Path(repo.get_commit_worktree(commit=commit).directory)

            return_code = await asyncio.to_thread(_execute, worktree_directory, repository, proposed_change)
            log.info(msg=f"repository_tests_completed return_code={return_code}")



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Any, Optional

from infrahub.core import registry
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.query.ipam import get_subnets
from infrahub.core.query.resource_manager import (
    PrefixPoolGetReserved,
    PrefixPoolSetReserved,
)
from infrahub.exceptions import ValidationError
from infrahub.pools.prefix import PrefixPool

from .. import Node

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.ipam.constants import IPNetworkType
    from infrahub.database import InfrahubDatabase


class CoreIPPrefixPool(Node):
    async def get_resource(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        identifier: Optional[str] = None,
        data: Optional[dict[str, Any]] = None,
        prefixlen: Optional[int] = None,
        member_type: Optional[str] = None,
        prefix_type: Optional[str] = None,
    ) -> Node:
        # Check if there is already a resource allocated with this identifier
        # if not, pull all existing prefixes and allocated the next available
        if identifier:
            query_get = await PrefixPoolGetReserved.init(db=db, pool_id=self.id, identifier=identifier)
            await query_get.execute(db=db)
            result = query_get.get_result()
            if result:
                prefix = result.get_node("prefix")
                # TODO add support for branch, if the node is reserved with this id in another branch we should return an error
                node = await registry.manager.get_one(db=db, id=prefix.get("uuid"), branch=branch)
                if node:
                    return node

        ip_namespace = await self.ip_namespace.get_peer(db=db)  # type: ignore[attr-defined]

        data = data or {}

        prefixlen = prefixlen or data.get("prefixlen", None) or self.default_prefix_length.value  # type: ignore[attr-defined]
        if not prefixlen:
            raise ValueError(
                f"IPPrefixPool: {self.name.value} | "  # type: ignore[attr-defined]
                "A prefixlen or a default_value must be provided to allocate a new prefix"
            )

        next_prefix = await self.get_next(db=db, prefixlen=prefixlen)

        prefix_type = prefix_type or data.get("prefix_type", None) or self.default_prefix_type.value  # type: ignore[attr-defined]
        if not prefix_type:
            raise ValueError(
                f"IPPrefixPool: {self.name.value} | "  # type: ignore[attr-defined]
                "A prefix_type or a default_value type must be provided to allocate a new prefix"
            )

        member_type = member_type or data.get("member_type", None) or self.default_member_type.value.value  # type: ignore[attr-defined]

        target_schema = registry.get_node_schema(name=prefix_type, branch=branch)
        node = await Node.init(db=db, schema=target_schema, branch=branch)
        try:
            await node.new(db=db, prefix=str(next_prefix), member_type=member_type, ip_namespace=ip_namespace, **data)
        except ValidationError as exc:
            raise ValueError(f"IPPrefixPool: {self.name.value} | {exc!s}") from exc  # type: ignore[attr-defined]
        await node.save(db=db)
        reconciler = IpamReconciler(db=db, branch=branch)
        await reconciler.reconcile(ip_value=next_prefix, namespace=ip_namespace.id, node_uuid=node.get_id())

        if identifier:
            query_set = await PrefixPoolSetReserved.init(
                db=db, pool_id=self.id, identifier=identifier, prefix_id=node.id
            )
            await query_set.execute(db=db)

        return node

    async def get_next(self, db: InfrahubDatabase, prefixlen: int) -> IPNetworkType:
        # Measure utilization of all prefixes identified as resources
        resources = await self.resources.get_peers(db=db)  # type: ignore[attr-defined]
        ip_namespace = await self.ip_namespace.get_peer(db=db)  # type: ignore[attr-defined]

        for resource in resources.values():
            subnets = await get_subnets(
                db=db,
                ip_prefix=ipaddress.ip_network(resource.prefix.value),  # type: ignore[attr-defined]
                namespace=ip_namespace,
                branch=self._branch,
                branch_agnostic=True,
            )

            pool = PrefixPool(resource.prefix.value)  # type: ignore[attr-defined]
            for subnet in subnets:
                pool.reserve(subnet=str(subnet.prefix))

            try:
                next_available = pool.get(prefixlen=prefixlen)
                return next_available
            except IndexError:
                continue

        raise IndexError("No more resources available")



from __future__ import annotations

import ipaddress
from collections import OrderedDict, defaultdict
from ipaddress import IPv4Network, IPv6Network
from typing import Optional, Union


class PrefixPool:
    """
    Class to automatically manage Prefixes and help to carve out sub-prefixes
    """

    def __init__(self, network: str) -> None:
        self.network = ipaddress.ip_network(network)

        # Define biggest and smallest possible masks
        self.mask_biggest = self.network.prefixlen + 1
        if self.network.version == 4:
            self.mask_smallest = 32
        else:
            self.mask_smallest = 128

        self.available_subnets = defaultdict(list)
        self.sub_by_key: dict[str, Optional[str]] = OrderedDict()
        self.sub_by_id: dict[str, str] = OrderedDict()

        # Save the top level available subnet
        for subnet in list(self.network.subnets(new_prefix=self.mask_biggest)):
            self.available_subnets[self.mask_biggest].append(str(subnet))

    def reserve(self, subnet: str, identifier: Optional[str] = None) -> bool:
        """
        Indicate that a specific subnet is already reserved/used
        """

        # TODO Add check to make sure the subnet provided has the right size
        sub = ipaddress.ip_network(subnet)

        if int(sub.prefixlen) <= int(self.network.prefixlen):
            raise ValueError(f"{subnet} do not have the right size ({sub.prefixlen},{self.network.prefixlen})")

        if sub.supernet(new_prefix=self.network.prefixlen) != self.network:
            raise ValueError(f"{subnet} is not part of this network")

        # Check first if this ID as already done a reservation
        if identifier and identifier in self.sub_by_id.keys():
            if self.sub_by_id[identifier] == str(sub):
                return True
            raise ValueError(
                f"this identifier ({identifier}) is already used but for a different resource ({self.sub_by_id[identifier]})"
            )

        if identifier and str(sub) in self.sub_by_key.keys():
            raise ValueError(f"this subnet is already reserved but not with this identifier ({identifier})")

        if str(sub) in self.sub_by_key.keys():
            self.remove_subnet_from_available_list(sub)
            return True

        # Check if the subnet itself is available
        # if available reserve and return
        if subnet in self.available_subnets[sub.prefixlen]:
            if identifier:
                self.sub_by_id[identifier] = subnet
                self.sub_by_key[subnet] = identifier
            else:
                self.sub_by_key[subnet] = None

            self.remove_subnet_from_available_list(sub)
            return True

        # If not reserved already, check if the subnet is available
        # start at sublen and check all available subnet
        # increase 1 by 1 until we find the closer supernet available
        # break it down and keep track of the other available subnets

        for sublen in range(sub.prefixlen - 1, self.network.prefixlen, -1):
            supernet = sub.supernet(new_prefix=sublen)
            if str(supernet) in self.available_subnets[sublen]:
                self.split_supernet(supernet=supernet, subnet=sub)
                return self.reserve(subnet=subnet, identifier=identifier)

        return False

    def get(self, prefixlen: int, identifier: Optional[str] = None) -> Union[IPv4Network, IPv6Network]:
        """Return the next available Subnet."""

        clean_prefixlen = int(prefixlen)

        if identifier and identifier in self.sub_by_id.keys():
            net = ipaddress.ip_network(self.sub_by_id[identifier])
            if net.prefixlen == clean_prefixlen:
                return net
            raise ValueError()

        if len(self.available_subnets[clean_prefixlen]) != 0:
            sub = self.available_subnets[clean_prefixlen][0]
            self.reserve(subnet=sub, identifier=identifier)
            return ipaddress.ip_network(sub)

        # if a subnet of this size is not available
        # we need to find the closest subnet available and split it
        for i in range(clean_prefixlen - 1, self.mask_biggest - 1, -1):
            if len(self.available_subnets[i]) != 0:
                supernet = ipaddress.ip_network(self.available_subnets[i][0])
                # supernet available, will split it
                subs = supernet.subnets(new_prefix=clean_prefixlen)
                next_sub: Union[IPv4Network, IPv6Network] = next(subs)  # type: ignore[assignment]
                self.split_supernet(supernet=supernet, subnet=next_sub)
                self.reserve(subnet=str(next_sub), identifier=identifier)
                return next_sub

        raise IndexError("No More subnet available")

    def get_nbr_available_subnets(self) -> dict[int, int]:
        tmp = {}
        for i in range(self.mask_biggest, self.mask_smallest + 1):
            tmp[i] = len(self.available_subnets[i])

        return tmp

    def check_if_already_allocated(self, identifier: str) -> bool:
        """
        Check if a subnet has already been allocated based on an identifier

        Need to add the same capability based on Network address
        If both identifier and subnet are provided, identifier take precedence
        """
        if identifier in self.sub_by_id.keys():
            return True
        return False

    def split_supernet(
        self, supernet: Union[IPv4Network, IPv6Network], subnet: Union[IPv4Network, IPv6Network]
    ) -> None:
        """Split a supernet into smaller networks"""

        # TODO ensure subnet is small than supernet
        # TODO ensure that subnet is part of supernet
        parent_net = supernet
        for i in range(supernet.prefixlen + 1, subnet.prefixlen + 1):
            tmp_net: list[Union[IPv4Network, IPv6Network]] = list(parent_net.subnets(new_prefix=i))

            if i == subnet.prefixlen:
                for net in tmp_net:
                    self.available_subnets[i].append(str(net))
            else:
                if subnet.subnet_of(other=tmp_net[0]):  # type: ignore[arg-type]
                    parent = 0
                    other = 1
                else:
                    parent = 1
                    other = 0

                parent_net = tmp_net[parent]
                self.available_subnets[i].append(str(tmp_net[other]))

        self.remove_subnet_from_available_list(supernet)

    def remove_subnet_from_available_list(self, subnet: Union[IPv4Network, IPv6Network]) -> None:
        """Remove a subnet from the list of available Subnet."""
        try:
            idx = self.available_subnets[subnet.prefixlen].index(str(subnet))
            del self.available_subnets[subnet.prefixlen][idx]
        except ValueError:
            # Already removed
            pass

        # if idx:
        #     return True
        # except:
        #     log.warn("Unable to remove %s from list of available subnets" % str(subnet))
        #     return False



from __future__ import annotations

import inspect
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generator, Optional, Union

from infrahub_sdk.uuidt import UUIDT

from infrahub.core.constants import RelationshipDirection, RelationshipStatus
from infrahub.core.query import Query, QueryType
from infrahub.core.query.subquery import build_subquery_filter, build_subquery_order
from infrahub.core.timestamp import Timestamp
from infrahub.core.utils import extract_field_filters

if TYPE_CHECKING:
    from uuid import UUID

    from neo4j.graph import Relationship as Neo4jRelationship

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.relationship import Relationship
    from infrahub.core.schema import RelationshipSchema
    from infrahub.database import InfrahubDatabase

# pylint: disable=redefined-builtin


@dataclass
class RelData:
    """Represent a relationship object in the database."""

    db_id: str
    branch: str
    type: str
    status: str

    @classmethod
    def from_db(cls, obj: Neo4jRelationship) -> RelData:
        return cls(db_id=obj.element_id, branch=obj.get("branch"), type=obj.type, status=obj.get("status"))


@dataclass
class FlagPropertyData:
    name: str
    prop_db_id: str
    rel: RelData
    value: bool


@dataclass
class NodePropertyData:
    name: str
    prop_db_id: str
    rel: RelData
    value: UUID


@dataclass
class RelationshipPeerData:
    branch: str

    source_id: UUID
    """UUID of the Source Node."""

    peer_id: UUID
    """UUID of the Peer Node."""

    peer_kind: str
    """Kind of the Peer Node."""

    properties: dict[str, Union[FlagPropertyData, NodePropertyData]]
    """UUID of the Relationship Node."""

    rel_node_id: Optional[UUID] = None
    """UUID of the Relationship Node."""

    peer_db_id: Optional[str] = None
    """Internal DB ID of the Peer Node."""

    rel_node_db_id: Optional[str] = None
    """Internal DB ID of the Relationship Node."""

    rels: Optional[list[RelData]] = None
    """Both relationships pointing at this Relationship Node."""

    updated_at: Optional[str] = None

    def rel_ids_per_branch(self) -> dict[str, list[Union[str, int]]]:
        response = defaultdict(list)
        for rel in self.rels:
            response[rel.branch].append(rel.db_id)

        for prop in self.properties.values():
            response[prop.rel.branch].append(prop.rel.db_id)

        return response


@dataclass
class RelationshipPeersData:
    id: UUID
    identifier: str
    source_id: UUID
    source_kind: str
    destination_id: UUID
    destination_kind: str

    def reversed(self) -> RelationshipPeersData:
        return RelationshipPeersData(
            id=self.id,
            identifier=self.identifier,
            source_id=self.destination_id,
            source_kind=self.destination_kind,
            destination_id=self.source_id,
            destination_kind=self.source_kind,
        )


@dataclass
class FullRelationshipIdentifier:
    identifier: str
    source_kind: str
    destination_kind: str


class RelationshipQuery(Query):
    def __init__(
        self,
        rel: Union[type[Relationship], Relationship] | None = None,
        rel_type: Optional[str] = None,
        source: Node | None = None,
        source_id: UUID | None = None,
        destination: Node | None = None,
        destination_id: UUID | None = None,
        schema: RelationshipSchema | None = None,
        branch: Branch | None = None,
        at: Union[Timestamp, str] | None = None,
        **kwargs,
    ):
        if not source and not source_id:
            raise ValueError("Either source or source_id must be provided.")
        if not rel and not rel_type:
            raise ValueError("Either rel or rel_type must be provided.")
        if not inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Rel must be a Relationship class or an instance of Relationship.")
        if not schema and inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Either an instance of Relationship or a valid schema must be provided.")

        self.source_id = source_id or source.id
        self.source = source

        # Destination is optional because not all RelationshipQuery needs it
        # If a query must have a destination defined, the validation must be done in the query specific init
        self.destination = destination
        self.destination_id = destination_id
        if not self.destination_id and destination:
            self.destination_id = destination.id

        self.rel = rel
        self.rel_type = rel_type or self.rel.rel_type
        self.schema = schema or self.rel.schema

        if not branch and inspect.isclass(rel) and not hasattr(rel, "branch"):
            raise ValueError("Either an instance of Relationship or a valid branch must be provided.")

        self.branch = branch or self.rel.branch

        if at:
            self.at = Timestamp(at)
        elif inspect.isclass(rel) and hasattr(rel, "at"):
            self.at = self.rel.at
        else:
            self.at = Timestamp()

        super().__init__(**kwargs)

    def get_relationship_properties_dict(self, status: RelationshipStatus) -> dict[str, Optional[str]]:
        rel_prop_dict = {
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "status": status.value,
            "from": self.at.to_string(),
        }
        if self.schema.hierarchical:
            rel_prop_dict["hierarchy"] = self.schema.hierarchical
        return rel_prop_dict


class RelationshipCreateQuery(RelationshipQuery):
    name = "relationship_create"

    type: QueryType = QueryType.WRITE

    def __init__(
        self,
        destination: Node = None,
        destination_id: UUID | None = None,
        **kwargs,
    ):
        if not destination and not destination_id:
            raise ValueError("Either destination or destination_id must be provided.")

        super().__init__(destination=destination, destination_id=destination_id, **kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.destination_id
        self.params["name"] = self.schema.identifier
        self.params["branch_support"] = self.schema.branch.value

        self.params["uuid"] = str(UUIDT())

        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()

        self.params["is_protected"] = self.rel.is_protected
        self.params["is_visible"] = self.rel.is_visible

        query_match = """
        MATCH (s:Node { uuid: $source_id })
        MATCH (d:Node { uuid: $destination_id })
        """
        self.add_to_query(query_match)

        self.query_add_all_node_property_match()

        self.params["rel_prop"] = self.get_relationship_properties_dict(status=RelationshipStatus.ACTIVE)
        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel_type} $rel_prop ]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel_type} $rel_prop ]{arrows.right.end}"

        query_create = """
        CREATE (rl:Relationship { uuid: $uuid, name: $name, branch_support: $branch_support })
        CREATE (s)%s(rl)
        CREATE (rl)%s(d)
        MERGE (ip:Boolean { value: $is_protected })
        MERGE (iv:Boolean { value: $is_visible })
        CREATE (rl)-[r3:IS_PROTECTED $rel_prop ]->(ip)
        CREATE (rl)-[r4:IS_VISIBLE $rel_prop ]->(iv)
        """ % (
            r1,
            r2,
        )

        self.add_to_query(query_create)
        self.return_labels = ["s", "d", "rl", "r1", "r2", "r3", "r4"]
        self.query_add_all_node_property_create()

    def query_add_all_node_property_match(self) -> None:
        for prop_name in self.rel._node_properties:
            if hasattr(self.rel, f"{prop_name}_id") and getattr(self.rel, f"{prop_name}_id"):
                self.query_add_node_property_match(name=prop_name)

    def query_add_node_property_match(self, name: str) -> None:
        self.add_to_query("MATCH (%s { uuid: $prop_%s_id })" % (name, name))
        self.params[f"prop_{name}_id"] = getattr(self.rel, f"{name}_id")
        self.return_labels.append(name)

    def query_add_all_node_property_create(self) -> None:
        for prop_name in self.rel._node_properties:
            if hasattr(self.rel, f"{prop_name}_id") and getattr(self.rel, f"{prop_name}_id"):
                self.query_add_node_property_create(name=prop_name)

    def query_add_node_property_create(self, name: str) -> None:
        query = """
        CREATE (rl)-[:HAS_%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(%s)
        """ % (
            name.upper(),
            name,
        )
        self.add_to_query(query)


class RelationshipUpdatePropertyQuery(RelationshipQuery):
    name = "relationship_property_update"
    type = QueryType.WRITE

    def __init__(
        self,
        properties_to_update: list[str],
        data: RelationshipPeerData,
        **kwargs,
    ):
        self.properties_to_update = properties_to_update
        self.data = data

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["rel_node_id"] = self.data.rel_node_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()

        query = """
        MATCH (rl:Relationship { uuid: $rel_node_id })
        """
        self.add_to_query(query)

        self.query_add_all_flag_property_merge()
        self.query_add_all_node_property_merge()

        self.query_add_all_flag_property_create()
        self.query_add_all_node_property_create()

    def query_add_all_flag_property_merge(self) -> None:
        for prop_name in self.rel._flag_properties:
            if prop_name in self.properties_to_update:
                self.query_add_flag_property_merge(name=prop_name)

    def query_add_flag_property_merge(self, name: str) -> None:
        self.add_to_query("MERGE (prop_%s:Boolean { value: $prop_%s })" % (name, name))
        self.params[f"prop_{name}"] = getattr(self.rel, name)
        self.return_labels.append(f"prop_{name}")

    def query_add_all_node_property_merge(self) -> None:
        for prop_name in self.rel._node_properties:
            if prop_name in self.properties_to_update:
                self.query_add_node_property_merge(name=prop_name)

    def query_add_node_property_merge(self, name: str) -> None:
        self.add_to_query("MERGE (prop_%s:Node { uuid: $prop_%s })" % (name, name))
        self.params[f"prop_{name}"] = getattr(self.rel, f"{name}_id")
        self.return_labels.append(f"prop_{name}")

    def query_add_all_flag_property_create(self) -> None:
        for prop_name in self.rel._flag_properties:
            if prop_name in self.properties_to_update:
                self.query_add_flag_property_create(name=prop_name)

    def query_add_flag_property_create(self, name: str) -> None:
        query = """
        CREATE (rl)-[:%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(prop_%s)
        """ % (
            name.upper(),
            name,
        )
        self.add_to_query(query)

    def query_add_all_node_property_create(self) -> None:
        for prop_name in self.rel._node_properties:
            if prop_name in self.properties_to_update:
                self.query_add_node_property_create(name=prop_name)

    def query_add_node_property_create(self, name: str) -> None:
        query = """
        CREATE (rl)-[:%s { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(prop_%s)
        """ % (
            "HAS_" + name.upper(),
            name,
        )
        self.add_to_query(query)


class RelationshipDataDeleteQuery(RelationshipQuery):
    name = "relationship_data_delete"
    type = QueryType.WRITE

    def __init__(
        self,
        data: RelationshipPeerData,
        **kwargs,
    ):
        self.data = data
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.data.peer_id
        self.params["rel_node_id"] = self.data.rel_node_id
        self.params["name"] = self.schema.identifier
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = self.at.to_string()

        # -----------------------------------------------------------------------
        # Match all nodes, including properties
        # -----------------------------------------------------------------------
        query = """
        MATCH (s:Node { uuid: $source_id })
        MATCH (d:Node { uuid: $destination_id })
        MATCH (rl:Relationship { uuid: $rel_node_id })
        """
        self.add_to_query(query)
        self.return_labels = ["s", "d", "rl"]

        for prop_name, prop in self.data.properties.items():
            self.add_to_query(
                "MATCH (prop_%(prop_name)s) WHERE %(id_func)s(prop_%(prop_name)s) = $prop_%(prop_name)s_id"
                % {"prop_name": prop_name, "id_func": db.get_id_function_name()}
            )
            self.params[f"prop_{prop_name}_id"] = db.to_database_id(prop.prop_db_id)
            self.return_labels.append(f"prop_{prop_name}")

        self.params["rel_prop"] = self.get_relationship_properties_dict(status=RelationshipStatus.DELETED)

        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel_type} $rel_prop ]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel_type} $rel_prop ]{arrows.right.end}"

        # -----------------------------------------------------------------------
        # Create all the DELETE relationships, including properties
        # -----------------------------------------------------------------------
        query = """
        CREATE (s)%s(rl)
        CREATE (rl)%s(d)
        """ % (
            r1,
            r2,
        )
        self.add_to_query(query)
        self.return_labels.extend(["r1", "r2"])

        for prop_name, prop in self.data.properties.items():
            self.add_to_query(
                "CREATE (prop_%s)<-[rel_prop_%s:%s $rel_prop ]-(rl)" % (prop_name, prop_name, prop.rel.type),
            )
            self.return_labels.append(f"rel_prop_{prop_name}")


class RelationshipDeleteQuery(RelationshipQuery):
    name = "relationship_delete"
    type = QueryType.WRITE

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if inspect.isclass(self.rel):
            raise TypeError("An instance of Relationship must be provided to RelationshipDeleteQuery")

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        rel_filter, rel_params = self.branch.get_query_filter_path(at=self.at, variable_name="edge")
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.destination_id
        self.params["rel_id"] = self.rel.id
        self.params["branch"] = self.branch.name
        self.params["rel_prop"] = self.get_relationship_properties_dict(status=RelationshipStatus.DELETED)
        self.params["at"] = self.at.to_string()
        self.params.update(rel_params)

        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel_type} $rel_prop ]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel_type} $rel_prop ]{arrows.right.end}"

        query = """
        MATCH (s:Node { uuid: $source_id })-[:IS_RELATED]-(rl:Relationship {uuid: $rel_id})-[:IS_RELATED]-(d:Node { uuid: $destination_id })
        WITH s, rl, d
        LIMIT 1
        CREATE (s)%(r1)s(rl)
        CREATE (rl)%(r2)s(d)
        WITH rl
        CALL {
            WITH rl
            MATCH (rl)-[edge:IS_VISIBLE]->(visible)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, visible
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:IS_VISIBLE $rel_prop]->(visible)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        CALL {
            WITH rl
            MATCH (rl)-[edge:IS_PROTECTED]->(protected)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, protected
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:IS_PROTECTED $rel_prop]->(protected)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        CALL {
            WITH rl
            MATCH (rl)-[edge:HAS_OWNER]->(owner_node)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, owner_node
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:HAS_OWNER $rel_prop]->(owner_node)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        CALL {
            WITH rl
            MATCH (rl)-[edge:HAS_SOURCE]->(source_node)
            WHERE %(rel_filter)s AND edge.status = "active"
            WITH rl, edge, source_node
            ORDER BY edge.branch_level DESC
            LIMIT 1
            CREATE (rl)-[deleted_edge:HAS_SOURCE $rel_prop]->(source_node)
            WITH edge
            WHERE edge.branch = $branch
            SET edge.to = $at
        }
        """ % {"r1": r1, "r2": r2, "rel_filter": rel_filter}

        self.params["at"] = self.at.to_string()
        self.return_labels = ["rl"]

        self.add_to_query(query)


class RelationshipGetPeerQuery(Query):
    name = "relationship_get_peer"
    type = QueryType.READ

    def __init__(
        self,
        filters: Optional[dict] = None,
        source: Optional[Node] = None,
        source_ids: Optional[list[str]] = None,
        source_kind: Optional[str] = None,
        rel: Optional[Union[type[Relationship], Relationship]] = None,
        rel_type: Optional[str] = None,
        schema: Optional[RelationshipSchema] = None,
        branch: Optional[Branch] = None,
        at: Optional[Union[Timestamp, str]] = None,
        **kwargs,
    ):
        if not source and not source_ids:
            raise ValueError("Either source or source_ids must be provided.")
        if not rel and not rel_type:
            raise ValueError("Either rel or rel_type must be provided.")
        if rel and not inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Rel must be a Relationship class or an instance of Relationship.")
        if not schema and inspect.isclass(rel) and not hasattr(rel, "schema"):
            raise ValueError("Either an instance of Relationship or a valid schema must be provided.")

        self.filters = filters or {}
        self.source_ids = source_ids or [source.id]
        self.source = source

        self.source_kind = source_kind or "Node"
        if source and not source_kind:
            self.source_kind = source.get_kind()

        self.rel = rel
        self.rel_type = rel_type or self.rel.rel_type
        self.schema = schema or self.rel.schema

        if not branch and inspect.isclass(rel) and not hasattr(rel, "branch"):
            raise ValueError("Either an instance of Relationship or a valid branch must be provided.")

        self.branch = branch or self.rel.branch

        if not at and inspect.isclass(rel) and hasattr(rel, "at"):
            self.at = self.rel.at
        else:
            self.at = Timestamp(at)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # pylint: disable=too-many-statements
        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at, branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)
        self.order_by = []

        peer_schema = self.schema.get_peer_schema(db=db, branch=self.branch)

        self.params["source_ids"] = self.source_ids
        self.params["rel_identifier"] = self.schema.identifier
        self.params["peer_kind"] = self.schema.peer
        self.params["source_kind"] = self.source_kind

        arrows = self.schema.get_query_arrows()

        path_str = (
            f"{arrows.left.start}[:IS_RELATED]{arrows.left.end}(rl){arrows.right.start}[:IS_RELATED]{arrows.right.end}"
        )

        branch_level_str = "reduce(br_lvl = 0, r in relationships(path) | br_lvl + r.branch_level)"
        froms_str = db.render_list_comprehension(items="relationships(path)", item_name="from")
        query = """
        MATCH (source_node:Node)%(arrow_left_start)s[:IS_RELATED]%(arrow_left_end)s(rl:Relationship { name: $rel_identifier })
        WHERE source_node.uuid IN $source_ids
        CALL {
            WITH rl, source_node
            MATCH path = (source_node)%(path)s(peer:Node)
            WHERE
                $source_kind IN LABELS(source_node) AND
                peer.uuid <> source_node.uuid AND
                $peer_kind IN LABELS(peer) AND
                all(r IN relationships(path) WHERE (%(branch_filter)s))
            WITH source_node, peer, rl, relationships(path) as rels, %(branch_level)s AS branch_level, %(froms)s AS froms
            RETURN peer as peer, rels, rl as rl1
            ORDER BY branch_level DESC, froms[-1] DESC, froms[-2] DESC
            LIMIT 1
        }
        WITH peer, rl1 as rl, rels, source_node
        """ % {
            "path": path_str,
            "branch_filter": branch_filter,
            "branch_level": branch_level_str,
            "froms": froms_str,
            "arrow_left_start": arrows.left.start,
            "arrow_left_end": arrows.left.end,
        }

        self.add_to_query(query)
        where_clause = ['all(r IN rels WHERE r.status = "active")']
        clean_filters = extract_field_filters(field_name=self.schema.name, filters=self.filters)

        if clean_filters and "id" in clean_filters or "ids" in clean_filters:
            where_clause.append("peer.uuid IN $peer_ids")
            self.params["peer_ids"] = clean_filters.get("ids", [])
            if clean_filters.get("id", None):
                self.params["peer_ids"].append(clean_filters.get("id"))

        self.add_to_query("WHERE " + " AND ".join(where_clause))

        self.return_labels = ["rl", "peer", "rels", "source_node"]

        # ----------------------------------------------------------------------------
        # FILTER Results
        # ----------------------------------------------------------------------------
        filter_cnt = 0
        for peer_filter_name, peer_filter_value in clean_filters.items():
            if "__" not in peer_filter_name:
                continue

            filter_cnt += 1

            filter_field_name, filter_next_name = peer_filter_name.split("__", maxsplit=1)

            if filter_field_name not in peer_schema.valid_input_names:
                continue

            field = peer_schema.get_field(filter_field_name)

            subquery, subquery_params, subquery_result_name = await build_subquery_filter(
                db=db,
                node_alias="peer",
                field=field,
                name=filter_field_name,
                filter_name=filter_next_name,
                filter_value=peer_filter_value,
                branch_filter=branch_filter,
                branch=self.branch,
                subquery_idx=filter_cnt,
            )
            self.params.update(subquery_params)

            with_str = ", ".join(
                [f"{subquery_result_name} as {label}" if label == "peer" else label for label in self.return_labels]
            )
            self.add_subquery(subquery=subquery, with_clause=with_str)

        # ----------------------------------------------------------------------------
        # QUERY Properties
        # ----------------------------------------------------------------------------
        query = """
        MATCH (rl)-[rel_is_visible:IS_VISIBLE]-(is_visible)
        MATCH (rl)-[rel_is_protected:IS_PROTECTED]-(is_protected)
        WHERE all(r IN [ rel_is_visible, rel_is_protected] WHERE (%s))
        """ % (branch_filter,)

        self.add_to_query(query)

        self.update_return_labels(["rel_is_visible", "rel_is_protected", "is_visible", "is_protected"])

        # Add Node Properties
        # We must query them one by one otherwise the second one won't return
        for node_prop in ["source", "owner"]:
            query = """
            WITH %s
            OPTIONAL MATCH (rl)-[rel_%s:HAS_%s]-(%s)
            WHERE all(r IN [ rel_%s ] WHERE (%s))
            """ % (
                ",".join(self.return_labels),
                node_prop,
                node_prop.upper(),
                node_prop,
                node_prop,
                branch_filter,
            )
            self.add_to_query(query)
            self.update_return_labels([f"rel_{node_prop}", node_prop])

        # ----------------------------------------------------------------------------
        # ORDER Results
        # ----------------------------------------------------------------------------
        if hasattr(peer_schema, "order_by") and peer_schema.order_by:
            order_cnt = 1

            for order_by_value in peer_schema.order_by:
                order_by_field_name, order_by_next_name = order_by_value.split("__", maxsplit=1)

                field = peer_schema.get_field(order_by_field_name)

                subquery, subquery_params, subquery_result_name = await build_subquery_order(
                    db=db,
                    field=field,
                    node_alias="peer",
                    name=order_by_field_name,
                    order_by=order_by_next_name,
                    branch_filter=branch_filter,
                    branch=self.branch,
                    subquery_idx=order_cnt,
                )
                self.order_by.append(subquery_result_name)
                self.params.update(subquery_params)

                self.add_subquery(subquery=subquery)

                order_cnt += 1

        else:
            self.order_by.append("peer.uuid")

    def get_peer_ids(self) -> list[str]:
        """Return a list of UUID of nodes associated with this relationship."""

        return [peer.peer_id for peer in self.get_peers()]

    def get_peers(self) -> Generator[RelationshipPeerData, None, None]:
        for result in self.get_results_group_by(("peer", "uuid"), ("source_node", "uuid")):
            rels = result.get("rels")
            data = RelationshipPeerData(
                source_id=result.get_node("source_node").get("uuid"),
                peer_id=result.get_node("peer").get("uuid"),
                peer_kind=result.get_node("peer").get("kind"),
                rel_node_db_id=result.get("rl").element_id,
                rel_node_id=result.get("rl").get("uuid"),
                updated_at=rels[0]["from"],
                rels=[RelData.from_db(rel) for rel in rels],
                branch=self.branch.name,
                properties={},
            )

            if hasattr(self.rel, "_flag_properties"):
                for prop in self.rel._flag_properties:
                    if prop_node := result.get(prop):
                        data.properties[prop] = FlagPropertyData(
                            name=prop,
                            prop_db_id=prop_node.element_id,
                            rel=RelData.from_db(result.get(f"rel_{prop}")),
                            value=prop_node.get("value"),
                        )

            if hasattr(self.rel, "_node_properties"):
                for prop in self.rel._node_properties:
                    if prop_node := result.get(prop):
                        data.properties[prop] = NodePropertyData(
                            name=prop,
                            prop_db_id=prop_node.element_id,
                            rel=RelData.from_db(result.get(f"rel_{prop}")),
                            value=prop_node.get("uuid"),
                        )

            yield data


class RelationshipGetQuery(RelationshipQuery):
    name = "relationship_get"

    type: QueryType = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["source_id"] = self.source_id
        self.params["destination_id"] = self.destination_id
        self.params["name"] = self.schema.identifier
        self.params["branch"] = self.branch.name

        rels_filter, rels_params = self.branch.get_query_filter_relationships(
            rel_labels=["r1", "r2"], at=self.at.to_string(), include_outside_parentheses=True
        )

        self.params.update(rels_params)

        arrows = self.schema.get_query_arrows()
        r1 = f"{arrows.left.start}[r1:{self.rel.rel_type}]{arrows.left.end}"
        r2 = f"{arrows.right.start}[r2:{self.rel.rel_type}]{arrows.right.end}"

        query = """
        MATCH (s:Node { uuid: $source_id })
        MATCH (d:Node { uuid: $destination_id })
        MATCH (s)%s(rl:Relationship { name: $name })%s(d)
        WHERE %s
        """ % (
            r1,
            r2,
            "\n AND ".join(rels_filter),
        )

        self.params["at"] = self.at.to_string()

        self.add_to_query(query)
        self.return_labels = ["s", "d", "rl", "r1", "r2"]


class RelationshipGetByIdentifierQuery(Query):
    name = "relationship_get_identifier"
    type = QueryType.READ

    def __init__(
        self,
        identifiers: Optional[list[str]] = None,
        full_identifiers: Optional[list[FullRelationshipIdentifier]] = None,
        excluded_namespaces: Optional[list[str]] = None,
        **kwargs,
    ) -> None:
        if (not identifiers and not full_identifiers) or (identifiers and full_identifiers):
            raise ValueError("one and only one of identifiers or full_identifiers is required")

        if full_identifiers:
            self.identifiers = list({i.identifier for i in full_identifiers})
            self.full_identifiers = full_identifiers
        else:
            self.identifiers = identifiers
            self.full_identifiers = []
        self.excluded_namespaces = excluded_namespaces or []

        # Always exclude relationships with internal nodes
        if "Internal" not in self.excluded_namespaces:
            self.excluded_namespaces.append("Internal")

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        self.params["identifiers"] = self.identifiers
        self.params["full_identifiers"] = [
            [full_id.source_kind, full_id.identifier, full_id.destination_kind] for full_id in self.full_identifiers
        ]
        self.params["excluded_namespaces"] = self.excluded_namespaces
        self.params["branch"] = self.branch.name
        self.params["at"] = self.at.to_string()

        rels_filter, rels_params = self.branch.get_query_filter_relationships(
            rel_labels=["r1", "r2"], at=self.at.to_string(), include_outside_parentheses=True
        )
        self.params.update(rels_params)

        query = """
        MATCH (rl:Relationship)
        WHERE rl.name IN $identifiers
        CALL {
            WITH rl
            MATCH (src:Node)-[r1:IS_RELATED]-(rl:Relationship)-[r2:IS_RELATED]-(dst:Node)
            WHERE (size($full_identifiers) = 0 OR [src.kind, rl.name, dst.kind] in $full_identifiers)
            AND NOT src.namespace IN $excluded_namespaces
            AND NOT dst.namespace IN $excluded_namespaces
            AND %s
            RETURN src, dst, r1, r2, rl as rl1
            ORDER BY r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC
            LIMIT 1
        }
        WITH src, dst, r1, r2, rl1 as rl
        WHERE r1.status = "active" AND r2.status = "active"
        """ % ("\n AND ".join(rels_filter),)

        self.add_to_query(query)
        self.return_labels = ["src", "dst", "rl"]

    def get_peers(self) -> Generator[RelationshipPeersData, None, None]:
        for result in self.get_results():
            data = RelationshipPeersData(
                id=result.get("rl").get("uuid"),
                identifier=result.get("rl").get("name"),
                source_id=result.get("src").get("uuid"),
                source_kind=result.get("src").get("kind"),
                destination_id=result.get("dst").get("uuid"),
                destination_kind=result.get("dst").get("kind"),
            )
            yield data


class RelationshipCountPerNodeQuery(Query):
    name = "relationship_count_per_node"
    type: QueryType = QueryType.READ

    def __init__(
        self,
        node_ids: list[str],
        identifier: str,
        direction: RelationshipDirection,
        **kwargs,
    ):
        self.node_ids = node_ids
        self.identifier = identifier
        self.direction = direction

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["peer_ids"] = self.node_ids
        self.params["rel_identifier"] = self.identifier

        path = "-[r:IS_RELATED]-"
        if self.direction == RelationshipDirection.OUTBOUND:
            path = "-[r:IS_RELATED]->"
        elif self.direction == RelationshipDirection.INBOUND:
            path = "<-[r:IS_RELATED]-"

        query = """
        MATCH (rl:Relationship { name: $rel_identifier })
        CALL {
            WITH rl
            MATCH path = (peer_node:Node)%(path)s(rl)
            WHERE peer_node.uuid IN $peer_ids AND %(branch_filter)s
            RETURN peer_node as peer, r as r1
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH peer as peer_node, r1 as r
        WHERE r.status = "active"
        """ % {"branch_filter": branch_filter, "path": path}

        self.add_to_query(query)
        self.return_labels = ["peer_node.uuid", "COUNT(peer_node.uuid) as nbr_peers"]

    async def get_count_per_peer(self) -> dict[str, int]:
        data: dict[str, int] = {}
        for result in self.results:
            data[result.get("peer_node.uuid")] = result.get("nbr_peers")

        for node_id in self.node_ids:
            if node_id not in data:
                data[node_id] = 0

        return data



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from prefect import get_client
from prefect.runtime import flow_run

from infrahub.core.registry import registry
from infrahub.tasks.registry import refresh_branches

from .constants import TAG_NAMESPACE, WorkflowTag

if TYPE_CHECKING:
    import logging

    from infrahub.services import InfrahubServices


async def add_tags(
    branches: list[str] | None = None,
    nodes: list[str] | None = None,
    others: list[str] | None = None,
    namespace: bool = True,
    db_change: bool = False,
) -> None:
    client = get_client(sync_client=False)
    current_flow_run_id = flow_run.id
    current_tags: list[str] = flow_run.tags
    branch_tags = [WorkflowTag.BRANCH.render(identifier=branch_name) for branch_name in branches] if branches else []
    node_tags = [WorkflowTag.RELATED_NODE.render(identifier=node_id) for node_id in nodes] if nodes else []
    others_tags = others or []
    new_tags = set(current_tags + branch_tags + node_tags + others_tags)
    if namespace:
        new_tags.add(TAG_NAMESPACE)
    if db_change:
        new_tags.add(WorkflowTag.DATABASE_CHANGE.render())
    await client.update_flow_run(current_flow_run_id, tags=list(new_tags))


async def add_branch_tag(branch_name: str) -> None:
    await add_tags(branches=[branch_name])


async def add_related_node_tag(node_id: str) -> None:
    await add_tags(nodes=[node_id])


async def wait_for_schema_to_converge(
    branch_name: str, service: InfrahubServices, log: logging.Logger | logging.LoggerAdapter
) -> None:
    has_converged = False
    branch_id = branch_name
    if branch := registry.branch.get(branch_name):
        branch_id = str(branch.get_uuid())

    delay = 0.2
    max_iterations = delay * 5 * 30
    iteration = 0
    while not has_converged:
        workers = await service.component.list_workers(branch=branch_id, schema_hash=True)

        hashes = {worker.schema_hash for worker in workers if worker.active}
        if len(hashes) == 1:
            has_converged = True
        else:
            await asyncio.sleep(delay=delay)

        if iteration >= max_iterations:
            log.warning(
                f"Schema had not converged after {delay * iteration:.2f} seconds, refreshing schema on local worker manually"
            )
            async with service.database.start_session() as db:
                await refresh_branches(db=db)
            return

        iteration += 1

    log.info(f"Schema converged after {delay * iteration} seconds")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from infrahub.core.constants import PathType
from infrahub.core.path import DataPath, GroupedDataPaths
from infrahub.core.schema import GenericSchema

from ..interface import ConstraintCheckerInterface
from ..shared import (
    RelationshipSchemaValidatorQuery,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class RelationshipPeerUpdateValidatorQuery(RelationshipSchemaValidatorQuery):
    name = "relationship_constraints_peer_validator"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        peer_schema = db.schema.get(name=self.relationship_schema.peer, branch=self.branch)
        allowed_peer_kinds = [peer_schema.kind]
        if isinstance(peer_schema, GenericSchema):
            allowed_peer_kinds += peer_schema.used_by

        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)

        self.params["relationship_id"] = self.relationship_schema.identifier
        self.params["allowed_peer_kinds"] = allowed_peer_kinds

        # ruff: noqa: E501
        query = """
        MATCH (n:%(node_kind)s)
        CALL {
            WITH n
            MATCH path = (root:Root)<-[rroot:IS_PART_OF]-(n)
            WHERE all(r in relationships(path) WHERE %(branch_filter)s)
            RETURN path as full_path, n as active_node
            ORDER BY rroot.branch_level DESC, rroot.from DESC
            LIMIT 1
        }
        WITH full_path, active_node
        WITH full_path, active_node
        WHERE all(r in relationships(full_path) WHERE r.status = "active")
        CALL {
            WITH active_node
            MATCH path = (active_node)-[rrel1:IS_RELATED]-(rel:Relationship { name: $relationship_id })-[rrel2:IS_RELATED]-(peer:Node)
            WHERE all(
                r in relationships(path)
                WHERE (%(branch_filter)s)
            )
            RETURN
                path as relationship_path,
                active_node as start_node,
                peer as peer_node,
                rrel1.branch_level + rrel2.branch_level AS branch_level_sum,
                [rrel1.from, rrel2.from] as from_times,
                // used as tiebreaker for updated relationships that were deleted and added at the same microsecond
                reduce(active_count = 0, r in relationships(path) | active_count + (CASE r.status WHEN "active" THEN 1 ELSE 0 END)) AS active_relationship_count,
                (CASE WHEN rrel1.branch_level > rrel2.branch_level THEN rrel1.branch ELSE rrel2.branch END) as deepest_branch_name
        }
        WITH
            collect([branch_level_sum, from_times, active_relationship_count, relationship_path, deepest_branch_name]) as enriched_paths,
            start_node,
            peer_node
        CALL {
            WITH enriched_paths, peer_node
            UNWIND enriched_paths as path_to_check
            RETURN path_to_check[3] as current_path, path_to_check[4] as branch_name, peer_node as current_peer
            ORDER BY
                path_to_check[0] DESC,
                path_to_check[1][1] DESC,
                path_to_check[1][0] DESC,
                path_to_check[2] DESC
            LIMIT 1
        }
        WITH start_node, current_peer, branch_name, current_path
        WITH start_node, current_peer, branch_name, current_path
        WHERE all(r in relationships(current_path) WHERE r.status = "active")
        AND NOT any(label IN LABELS(current_peer) WHERE label IN $allowed_peer_kinds)
        """ % {"branch_filter": branch_filter, "node_kind": self.node_schema.kind}

        self.add_to_query(query)
        self.return_labels = ["start_node.uuid", "branch_name", "current_peer.uuid"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("branch_name")),
                    path_type=PathType.NODE,
                    node_id=str(result.get("start_node.uuid")),
                    field_name=self.relationship_schema.name,
                    peer_id=str(result.get("current_peer.uuid")),
                    kind=self.node_schema.kind,
                )
            )

        return grouped_data_paths


class RelationshipPeerChecker(ConstraintCheckerInterface):
    query_classes = [RelationshipPeerUpdateValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Optional[Branch] = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "relationship.peer.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []

        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db, branch=self.branch, node_schema=request.node_schema, schema_path=request.schema_path
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.services import InfrahubServices


async def refresh_heartbeat(service: InfrahubServices) -> None:
    """Update API server worker information in the cache

    The goal of this job is to provide an updated list of API server workers in the cache, it will have a freshness
    of 15 seconds after which workers that haven't updated their entry in the cache will be removed.

    The function also keeps the primary API server ID up to date. The primary ID is used within other tasks to ensure
    that only one worker is responsible for scheduling specific tasks.
    """
    service.log.debug("Refreshing API workers in cache")
    await service.component.refresh_heartbeat()



import logging
import sys

import typer
from infrahub_sdk import Config, InfrahubClientSync

from infrahub import config
from infrahub.core.constants import InfrahubKind

logging.getLogger("httpx").setLevel(logging.ERROR)
app = typer.Typer()


def parse_helper_get_input(text: str) -> str:
    """Parse the input provided to infrahub-githelper get

    Format1 (with usehttppath true)
      protocol=https
      host=github.com
      path=opsmill/infrahub-demo-edge.git

    Format2 (without usehttppath true)(default)
      protocol=https
      host=github.com
    """
    input_dict = {line.split("=")[0]: line.split("=")[1] for line in text.splitlines() if "=" in line}

    if "protocol" not in input_dict or "host" not in input_dict:
        raise ValueError("Input format not supported.")
    if "path" not in input_dict:
        raise ValueError(
            """Git usehttppath must be enabled to use this helper. You can active it with
    git config --global credential.usehttppath true
    """
        )

    return f"{input_dict['protocol']}://{input_dict['host']}/{input_dict['path']}"


@app.command()
def get(
    input_str: str = typer.Argument(... if sys.stdin.isatty() else sys.stdin.read().strip()),
    config_file: str = typer.Option("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    config.SETTINGS.initialize_and_exit(config_file=config_file)

    try:
        location = parse_helper_get_input(text=input_str)
    except ValueError as exc:
        print(str(exc))
        raise typer.Exit(1) from exc

    client = InfrahubClientSync(config=Config(address=config.SETTINGS.main.internal_address, insert_tracker=True))
    repo = client.get(kind=InfrahubKind.GENERICREPOSITORY, location__value=location)

    if not repo:
        print("Repository not found in the database.")
        raise typer.Exit(1)

    if not repo.credential._id:
        print("Repository doesn't have credentials defined.")
        raise typer.Exit(1)

    repo.credential.fetch()

    print(f"username={repo.credential.peer.username.value}")
    print(f"password={repo.credential.peer.password.value}")


# pylint: disable=unused-argument
@app.command()
def store(
    input_str: str = typer.Argument(None),
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    raise typer.Exit()



from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .calculator import DiffCalculatorDependency
from .combiner import DiffCombinerDependency
from .conflict_transferer import DiffConflictTransfererDependency
from .conflicts_enricher import DiffConflictsEnricherDependency
from .data_check_synchronizer import DiffDataCheckSynchronizerDependency
from .enricher.aggregated import DiffAggregatedEnricherDependency
from .enricher.labels import DiffLabelsEnricherDependency
from .enricher.summary_counts import DiffSummaryCountsEnricherDependency
from .repository import DiffRepositoryDependency


class DiffCoordinatorDependency(DependencyBuilder[DiffCoordinator]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffCoordinator:
        return DiffCoordinator(
            diff_repo=DiffRepositoryDependency.build(context=context),
            diff_calculator=DiffCalculatorDependency.build(context=context),
            diff_combiner=DiffCombinerDependency.build(context=context),
            diff_enricher=DiffAggregatedEnricherDependency.build(context=context),
            conflicts_enricher=DiffConflictsEnricherDependency.build(context=context),
            labels_enricher=DiffLabelsEnricherDependency.build(context=context),
            summary_counts_enricher=DiffSummaryCountsEnricherDependency.build(context=context),
            data_check_synchronizer=DiffDataCheckSynchronizerDependency.build(context=context),
            conflict_transferer=DiffConflictTransfererDependency.build(context=context),
        )



from __future__ import annotations

from pathlib import Path

import typer
import yaml
from pydantic import ValidationError
from rich.console import Console

from infrahub_sdk.ctl.client import initialize_client

from ..async_typer import AsyncTyper
from ..ctl.exceptions import FileNotValidError
from ..ctl.utils import init_logging
from ..graphql import Mutation
from ..schema.repository import InfrahubRepositoryConfig
from ._file import read_file
from .parameters import CONFIG_PARAM

app = AsyncTyper()
console = Console()


def get_repository_config(repo_config_file: Path) -> InfrahubRepositoryConfig:
    try:
        config_file_data = load_repository_config_file(repo_config_file)
    except FileNotFoundError as exc:
        console.print(f"[red]File not found {exc}")
        raise typer.Exit(1) from exc
    except FileNotValidError as exc:
        console.print(f"[red]{exc.message}")
        raise typer.Exit(1) from exc

    try:
        data = InfrahubRepositoryConfig(**config_file_data)
    except ValidationError as exc:
        console.print(f"[red]Repository config file not valid, found {len(exc.errors())} error(s)")
        for error in exc.errors():
            loc_str = [str(item) for item in error["loc"]]
            console.print(f"  {'/'.join(loc_str)} | {error['msg']} ({error['type']})")
        raise typer.Exit(1) from exc

    return data


def load_repository_config_file(repo_config_file: Path) -> dict:
    yaml_data = read_file(file_name=repo_config_file)

    try:
        data = yaml.safe_load(yaml_data)
    except yaml.YAMLError as exc:
        raise FileNotValidError(name=str(repo_config_file)) from exc

    return data


@app.callback()
def callback() -> None:
    """
    Manage the repositories in a remote Infrahub instance.

    List, create, delete ..
    """


@app.command()
async def add(
    name: str,
    location: str,
    description: str = "",
    username: str | None = None,
    password: str = "",
    commit: str = "",
    read_only: bool = False,
    debug: bool = False,
    branch: str = typer.Option("main", help="Branch on which to add the repository."),
    _: str = CONFIG_PARAM,
) -> None:
    """Add a new repository."""

    init_logging(debug=debug)

    input_data = {
        "data": {
            "name": {"value": name},
            "location": {"value": location},
            "description": {"value": description},
            "commit": {"value": commit},
        },
    }

    client = initialize_client()

    credential = await client.create(kind="CorePasswordCredential", name=name, username=username, password=password)
    await credential.save(allow_upsert=True)
    input_data["data"]["credential"] = {"id": credential.id}

    query = Mutation(
        mutation="CoreReadOnlyRepositoryCreate" if read_only else "CoreRepositoryCreate",
        input_data=input_data,
        query={"ok": None},
    )

    await client.execute_graphql(query=query.render(), branch_name=branch, tracker="mutation-repository-create")



from pydantic import BaseModel, ConfigDict, Field

from infrahub.core.branch import Branch
from infrahub.core.models import SchemaUpdateMigrationInfo
from infrahub.core.path import SchemaPath
from infrahub.core.schema.schema_branch import SchemaBranch


class SchemaApplyMigrationData(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True, json_encoders={SchemaBranch: SchemaBranch.to_dict_schema_object}
    )
    branch: Branch
    new_schema: SchemaBranch
    previous_schema: SchemaBranch
    migrations: list[SchemaUpdateMigrationInfo]


class SchemaMigrationPathResponseData(BaseModel):
    errors: list[str] = Field(default_factory=list)
    migration_name: str | None = None
    nbr_migrations_executed: int | None = None
    schema_path: SchemaPath | None = None



from infrahub.core.protocols import CoreMenuItem
from infrahub.database import InfrahubDatabase

from .models import MenuItemDefinition


async def create_menu_children(db: InfrahubDatabase, parent: CoreMenuItem, children: list[MenuItemDefinition]) -> None:
    for child in children:
        obj = await child.to_node(db=db, parent=parent)
        await obj.save(db=db)
        if child.children:
            await create_menu_children(db=db, parent=obj, children=child.children)



from infrahub.core.ipam.kinds_getter import IpamKindsGetter
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class IpamKindsGetterDependency(DependencyBuilder[IpamKindsGetter]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> IpamKindsGetter:
        return IpamKindsGetter(db=context.db)



import pytest
from infrahub_testcontainers.helpers import TestInfrahubDocker

from .. import Config, InfrahubClient, InfrahubClientSync


class TestInfrahubDockerClient(TestInfrahubDocker):
    @pytest.fixture(scope="class")
    def client(self, infrahub_port: int) -> InfrahubClient:
        return InfrahubClient(
            config=Config(username="admin", password="infrahub", address=f"http://localhost:{infrahub_port}")  # noqa: S106
        )

    @pytest.fixture(scope="class")
    def client_sync(self, infrahub_port: int) -> InfrahubClientSync:
        return InfrahubClientSync(
            config=Config(username="admin", password="infrahub", address=f"http://localhost:{infrahub_port}")  # noqa: S106
        )



from graphql.type.definition import GraphQLArgument, GraphQLList
from graphql.type.directives import DirectiveLocation, GraphQLDirective, specified_directives
from graphql.type.scalars import GraphQLString

GraphQLExpand = GraphQLDirective(
    name="expand",
    locations=[
        DirectiveLocation.FIELD,
        DirectiveLocation.FRAGMENT_SPREAD,
        DirectiveLocation.INLINE_FRAGMENT,
    ],
    args={
        "exclude": GraphQLArgument(GraphQLList(GraphQLString), description="Exclude specific fields"),
    },
    description="Expands a field to include Node defaults",
)


DIRECTIVES = list(specified_directives) + [GraphQLExpand]



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventBranchDelete(InfrahubMessage):
    """Sent when a branch has been deleted."""

    branch: str = Field(..., description="The branch that was deleted")
    branch_id: str = Field(..., description="The unique ID of the branch")
    sync_with_git: bool = Field(..., description="Indicates if the branch was extended to Git")



from typing import Generator

from infrahub import config
from infrahub.core import registry
from infrahub.core.diff.query.field_summary import EnrichedDiffNodeFieldSummaryQuery
from infrahub.core.query.diff import DiffCountChanges
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import ResourceNotFoundError
from infrahub.log import get_logger

from ..model.path import (
    ConflictSelection,
    EnrichedDiffConflict,
    EnrichedDiffRoot,
    EnrichedDiffRootMetadata,
    EnrichedDiffs,
    EnrichedDiffsMetadata,
    EnrichedNodeCreateRequest,
    NodeDiffFieldSummary,
    NodeFieldSpecifier,
    TimeRange,
    TrackingId,
)
from ..query.delete_query import EnrichedDiffDeleteQuery
from ..query.diff_get import EnrichedDiffGetQuery
from ..query.diff_summary import DiffSummaryCounters, DiffSummaryQuery
from ..query.drop_tracking_id import EnrichedDiffDropTrackingIdQuery
from ..query.field_specifiers import EnrichedDiffFieldSpecifiersQuery
from ..query.filters import EnrichedDiffQueryFilters
from ..query.get_conflict_query import EnrichedDiffConflictQuery
from ..query.has_conflicts_query import EnrichedDiffHasConflictQuery
from ..query.roots_metadata import EnrichedDiffRootsMetadataQuery
from ..query.save import EnrichedDiffRootsCreateQuery, EnrichedNodeBatchCreateQuery, EnrichedNodesLinkQuery
from ..query.time_range_query import EnrichedDiffTimeRangeQuery
from ..query.update_conflict_query import EnrichedDiffConflictUpdateQuery
from .deserializer import EnrichedDiffDeserializer

log = get_logger()


class DiffRepository:
    MAX_SAVE_BATCH_SIZE: int = 100

    def __init__(self, db: InfrahubDatabase, deserializer: EnrichedDiffDeserializer):
        self.db = db
        self.deserializer = deserializer

    async def get(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        filters: dict | None = None,
        include_parents: bool = True,
        limit: int | None = None,
        offset: int | None = None,
        tracking_id: TrackingId | None = None,
        diff_ids: list[str] | None = None,
        include_empty: bool = False,
    ) -> list[EnrichedDiffRoot]:
        final_max_depth = config.SETTINGS.database.max_depth_search_hierarchy
        query = await EnrichedDiffGetQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_names=diff_branch_names,
            from_time=from_time,
            to_time=to_time,
            filters=EnrichedDiffQueryFilters(**dict(filters or {})),
            max_depth=final_max_depth,
            limit=limit,
            offset=offset,
            tracking_id=tracking_id,
            diff_ids=diff_ids,
        )
        await query.execute(db=self.db)
        diff_roots = await self.deserializer.deserialize(
            database_results=query.get_results(), include_parents=include_parents
        )
        if not include_empty:
            diff_roots = [dr for dr in diff_roots if len(dr.nodes) > 0]
        return diff_roots

    async def get_pairs(
        self,
        base_branch_name: str,
        diff_branch_name: str,
        from_time: Timestamp,
        to_time: Timestamp,
    ) -> list[EnrichedDiffs]:
        max_depth = config.SETTINGS.database.max_depth_search_hierarchy
        query = await EnrichedDiffGetQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_names=[diff_branch_name],
            from_time=from_time,
            to_time=to_time,
            max_depth=max_depth,
        )
        await query.execute(db=self.db)
        diff_branch_roots = await self.deserializer.deserialize(
            database_results=query.get_results(), include_parents=True
        )
        diffs_by_uuid = {dbr.uuid: dbr for dbr in diff_branch_roots}
        base_partner_query = await EnrichedDiffGetQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_names=[base_branch_name],
            max_depth=max_depth,
            diff_ids=[d.partner_uuid for d in diffs_by_uuid.values()],
        )
        await base_partner_query.execute(db=self.db)
        base_branch_roots = await self.deserializer.deserialize(
            database_results=base_partner_query.get_results(), include_parents=True
        )
        diffs_by_uuid.update({bbr.uuid: bbr for bbr in base_branch_roots})
        return [
            EnrichedDiffs(
                base_branch_name=base_branch_name,
                diff_branch_name=diff_branch_name,
                base_branch_diff=diffs_by_uuid[dbr.partner_uuid],
                diff_branch_diff=dbr,
            )
            for dbr in diff_branch_roots
        ]

    async def hydrate_diff_pair(self, enriched_diffs_metadata: EnrichedDiffsMetadata) -> EnrichedDiffs:
        hydrated_base_diff = await self.get_one(
            diff_branch_name=enriched_diffs_metadata.base_branch_name,
            diff_id=enriched_diffs_metadata.base_branch_diff.uuid,
        )
        hydrated_branch_diff = await self.get_one(
            diff_branch_name=enriched_diffs_metadata.diff_branch_name,
            diff_id=enriched_diffs_metadata.diff_branch_diff.uuid,
        )
        return EnrichedDiffs(
            base_branch_name=enriched_diffs_metadata.base_branch_name,
            diff_branch_name=enriched_diffs_metadata.diff_branch_name,
            base_branch_diff=hydrated_base_diff,
            diff_branch_diff=hydrated_branch_diff,
        )

    async def get_one(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        filters: dict | None = None,
        include_parents: bool = True,
    ) -> EnrichedDiffRoot:
        enriched_diffs = await self.get(
            base_branch_name=registry.default_branch,
            diff_branch_names=[diff_branch_name],
            tracking_id=tracking_id,
            diff_ids=[diff_id] if diff_id else None,
            filters=filters,
            include_parents=include_parents,
            include_empty=True,
        )
        error_str = f"branch {diff_branch_name}"
        if tracking_id:
            error_str += f" with tracking_id {tracking_id.serialize()}"
        if diff_id:
            error_str += f" with ID {diff_id}"
        if len(enriched_diffs) == 0:
            raise ResourceNotFoundError(f"Cannot find diff for {error_str}")
        if len(enriched_diffs) > 1:
            raise ResourceNotFoundError(f"Multiple diffs for {error_str}")
        return enriched_diffs[0]

    def _get_node_create_request_batch(
        self, enriched_diffs: EnrichedDiffs
    ) -> Generator[list[EnrichedNodeCreateRequest], None, None]:
        node_requests = []
        for diff_root in (enriched_diffs.base_branch_diff, enriched_diffs.diff_branch_diff):
            for node in diff_root.nodes:
                node_requests.append(EnrichedNodeCreateRequest(node=node, root_uuid=diff_root.uuid))
                if len(node_requests) == self.MAX_SAVE_BATCH_SIZE:
                    yield node_requests
                    node_requests = []
        if node_requests:
            yield node_requests

    @retry_db_transaction(name="enriched_diff_save")
    async def save(self, enriched_diffs: EnrichedDiffs) -> None:
        num_nodes = len(enriched_diffs.base_branch_diff.nodes) + len(enriched_diffs.diff_branch_diff.nodes)
        log.info(f"Saving diff (num_nodes={num_nodes})...")
        root_query = await EnrichedDiffRootsCreateQuery.init(db=self.db, enriched_diffs=enriched_diffs)
        await root_query.execute(db=self.db)
        for node_create_batch in self._get_node_create_request_batch(enriched_diffs=enriched_diffs):
            node_query = await EnrichedNodeBatchCreateQuery.init(db=self.db, node_create_batch=node_create_batch)
            await node_query.execute(db=self.db)
        link_query = await EnrichedNodesLinkQuery.init(db=self.db, enriched_diffs=enriched_diffs)
        await link_query.execute(db=self.db)
        log.info("Diff saved.")

    async def summary(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
        filters: dict | None = None,
    ) -> DiffSummaryCounters | None:
        query = await DiffSummaryQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_names=diff_branch_names,
            filters=EnrichedDiffQueryFilters(**dict(filters or {})),
            from_time=from_time,
            to_time=to_time,
            tracking_id=tracking_id,
        )
        await query.execute(db=self.db)
        return query.get_summary()

    async def delete_diff_roots(self, diff_root_uuids: list[str]) -> None:
        query = await EnrichedDiffDeleteQuery.init(db=self.db, enriched_diff_root_uuids=diff_root_uuids)
        await query.execute(db=self.db)

    async def get_time_ranges(
        self,
        base_branch_name: str,
        diff_branch_name: str,
        from_time: Timestamp,
        to_time: Timestamp,
    ) -> list[TimeRange]:
        query = await EnrichedDiffTimeRangeQuery.init(
            db=self.db,
            base_branch_name=base_branch_name,
            diff_branch_name=diff_branch_name,
            from_time=from_time,
            to_time=to_time,
        )
        await query.execute(db=self.db)
        return await query.get_time_ranges()

    async def get_diff_pairs_metadata(
        self,
        diff_branch_names: list[str] | None = None,
        base_branch_names: list[str] | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
    ) -> list[EnrichedDiffsMetadata]:
        if diff_branch_names and base_branch_names:
            diff_branch_names += base_branch_names
        empty_roots = await self.get_roots_metadata(
            diff_branch_names=diff_branch_names,
            base_branch_names=base_branch_names,
            from_time=from_time,
            to_time=to_time,
        )
        roots_by_id = {root.uuid: root for root in empty_roots}
        pairs: list[EnrichedDiffsMetadata] = []
        for branch_root in empty_roots:
            if branch_root.base_branch_name == branch_root.diff_branch_name:
                continue
            base_root = roots_by_id[branch_root.partner_uuid]
            pairs.append(
                EnrichedDiffsMetadata(
                    base_branch_name=branch_root.base_branch_name,
                    diff_branch_name=branch_root.diff_branch_name,
                    base_branch_diff=base_root,
                    diff_branch_diff=branch_root,
                )
            )
        return pairs

    async def get_roots_metadata(
        self,
        diff_branch_names: list[str] | None = None,
        base_branch_names: list[str] | None = None,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
    ) -> list[EnrichedDiffRootMetadata]:
        query = await EnrichedDiffRootsMetadataQuery.init(
            db=self.db,
            diff_branch_names=diff_branch_names,
            base_branch_names=base_branch_names,
            from_time=from_time,
            to_time=to_time,
        )
        await query.execute(db=self.db)
        diff_roots = []
        for neo4j_node in query.get_root_nodes_metadata():
            diff_roots.append(self.deserializer.build_diff_root_metadata(root_node=neo4j_node))
        return diff_roots

    async def diff_has_conflicts(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
    ) -> bool:
        query = await EnrichedDiffHasConflictQuery.init(
            db=self.db, diff_branch_name=diff_branch_name, tracking_id=tracking_id, diff_id=diff_id
        )
        await query.execute(db=self.db)
        return await query.has_conflict()

    async def get_conflict_by_id(self, conflict_id: str) -> EnrichedDiffConflict:
        query = await EnrichedDiffConflictQuery.init(db=self.db, conflict_id=conflict_id)
        await query.execute(db=self.db)
        conflict_node = await query.get_conflict_node()
        if not conflict_node:
            raise ResourceNotFoundError(f"No conflict with id {conflict_id}")
        return self.deserializer.deserialize_conflict(diff_conflict_node=conflict_node)

    async def update_conflict_by_id(
        self, conflict_id: str, selection: ConflictSelection | None
    ) -> EnrichedDiffConflict:
        query = await EnrichedDiffConflictUpdateQuery.init(db=self.db, conflict_id=conflict_id, selection=selection)
        await query.execute(db=self.db)
        conflict_node = await query.get_conflict_node()
        if not conflict_node:
            raise ResourceNotFoundError(f"No conflict with id {conflict_id}")
        return self.deserializer.deserialize_conflict(diff_conflict_node=conflict_node)

    async def get_node_field_summaries(
        self, diff_branch_name: str, tracking_id: TrackingId | None = None, diff_id: str | None = None
    ) -> list[NodeDiffFieldSummary]:
        query = await EnrichedDiffNodeFieldSummaryQuery.init(
            db=self.db, diff_branch_name=diff_branch_name, tracking_id=tracking_id, diff_id=diff_id
        )
        await query.execute(db=self.db)
        return await query.get_field_summaries()

    async def drop_tracking_ids(self, tracking_ids: list[TrackingId]) -> None:
        query = await EnrichedDiffDropTrackingIdQuery.init(db=self.db, tracking_ids=tracking_ids)
        await query.execute(db=self.db)

    async def get_num_changes_in_time_range_by_branch(
        self, branch_names: list[str], from_time: Timestamp, to_time: Timestamp
    ) -> dict[str, int]:
        query = await DiffCountChanges.init(db=self.db, branch_names=branch_names, diff_from=from_time, diff_to=to_time)
        await query.execute(db=self.db)
        return query.get_num_changes_by_branch()

    async def get_node_field_specifiers(self, diff_id: str) -> set[NodeFieldSpecifier]:
        limit = config.SETTINGS.database.query_size_limit
        offset = 0
        specifiers: set[NodeFieldSpecifier] = set()
        while True:
            query = await EnrichedDiffFieldSpecifiersQuery.init(db=self.db, diff_id=diff_id, offset=offset, limit=limit)
            await query.execute(db=self.db)

            new_specifiers = {
                NodeFieldSpecifier(
                    node_uuid=field_specifier_tuple[0],
                    field_name=field_specifier_tuple[1],
                )
                for field_specifier_tuple in query.get_node_field_specifier_tuples()
            }
            if not new_specifiers:
                break
            specifiers |= new_specifiers
            offset += limit
        return specifiers



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import ObjectType
from graphene.types.objecttype import ObjectTypeOptions

from infrahub import config

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext


class InfrahubObjectTypeOptions(ObjectTypeOptions):
    model = None


class InfrahubObjectType(ObjectType):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls, model=None, interfaces=(), _meta=None, **options
    ) -> None:
        if not _meta:
            _meta = InfrahubObjectTypeOptions(cls)

        _meta.model = model

        super().__init_subclass_with_meta__(_meta=_meta, interfaces=interfaces, **options)

    @classmethod
    async def get_list(cls, fields: dict[str, Any], context: GraphqlContext, **kwargs) -> list[dict[str, Any]]:
        async with context.db.session(database=config.SETTINGS.database.database_name) as db:
            filters = {key: value for key, value in kwargs.items() if "__" in key and value}

            if filters:
                objs = await cls._meta.model.get_list(
                    filters=filters,
                    at=context.at,
                    branch=context.branch,
                    account=context.account_session,
                    include_source=True,
                    db=db,
                )
            else:
                objs = await cls._meta.model.get_list(
                    at=context.at,
                    branch=context.branch,
                    account=context.account_session,
                    include_source=True,
                    db=db,
                )

            if not objs:
                return []

            return [obj.to_graphql(fields=fields) for obj in objs]



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventSchemaUpdate(InfrahubMessage):
    """Sent when the schema on a branch has been updated."""

    branch: str = Field(..., description="The branch where the update occurred")
